# def direct_dinf(grid, elevs='topographic_elevation', baselevel_nodes=None):


"""
flow_direction_dinf.py: calculate Dinfinity flow direction on raster grids.

Calculates flow direction and proportion on a raster grid by the Dinfinity
algorithm of Tarboton 1997.

KRB Feb 2017
"""

from landlab.core.utils import as_id_array
from landlab.utils.return_array import return_array_at_node
from landlab import BAD_INDEX_VALUE, CLOSED_BOUNDARY
UNDEFINED_INDEX = BAD_INDEX_VALUE
from landlab import VoronoiDelaunayGrid  # for type tests

import numpy as np


def flow_directions_dinf(grid,
                         elevs='topographic__elevation',
                         baselevel_nodes=None):
    """
    Find Dinfinity flow directions and proportions on a raster grid.

    Finds and returns flow directions and proportions for a given elevation
    grid by the D infinity method (Tarboton, 1997). Each node is assigned two
    flow directions, toward the two neighboring nodes that are on the steepest
    subtriangle. Partitioning of flow is done based on the aspect of the
    subtriangle.

    This method does not support irregular grids.

    Parameters
    ----------
    grid : ModelGrid
        A grid of type Voroni.
    elevs : field name at node or array of length node
        The surface to direct flow across.
    baselevel_nodes : array_like, optional
        IDs of open boundary (baselevel) nodes.

    Returns
    -------
    receivers : ndarray of size (num nodes, max neighbors at node)
        For each node, the IDs of the nodes that receive its flow. For nodes
        that do not direct flow to all neighbors, BAD_INDEX_VALUE is given as
        a placeholder. The ID of the node itself is given if no other receiver
        is assigned.
    proportions : ndarray of size (num nodes, max neighbors at node)
        For each receiver, the proportion of flow (between 0 and 1) is given.
        A proportion of zero indicates that the link does not have flow along
        it.
    steepest_slope : ndarray
        The slope value (positive downhill) in the direction of flow.
    steepest_receiver : ndarray
        For each node, the node ID of the node connected by the steepest link.
        BAD_INDEX_VALUE is given if no flow emmanates from the node.
    sink : ndarray
        IDs of nodes that are flow sinks (they are their own receivers)
    receiver_links : ndarray of size (num nodes, max neighbors at node)
        ID of links that leads from each node to its receiver, or
        UNDEFINED_INDEX if no flow occurs on this link.
    steepest_link : ndarray
        For each node, the link ID of the steepest link.
        BAD_INDEX_VALUE is given if no flow emmanates from the node.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.flow_director.flow_direction_dinf import(
    ...                                                  flow_directions_dinf)

    Dinfinity routes flow based on the relative proportion of flow along the
    triangular facets around a central raster node.

    >>> grid = RasterModelGrid((3,3), spacing=(1, 1))
    >>> _ = grid.add_field('topographic__elevation',
    ...                     2.*grid.node_x+grid.node_y,
    ...                     at = 'node')
    >>> (receivers, proportions,
    ... steepest_slope, steepest_receiver,
    ... sink, receiver_links, steepest_link) = flow_directions_dinf(grid)
    >>> receivers
    array([[ 0, -1],
           [ 0,  3],
           [ 1,  4],
           [ 0,  1],
           [ 3,  0],
           [ 4,  1],
           [ 3,  4],
           [ 6,  3],
           [ 7,  4]])
    >>> proportions
    array([[ 1.        ,  0.        ],
           [ 1.        , -0.        ],
           [ 1.        , -0.        ],
           [ 1.        ,  0.        ],
           [ 0.40966553,  0.59033447],
           [ 0.40966553,  0.59033447],
           [ 1.        ,  0.        ],
           [ 0.40966553,  0.59033447],
           [ 0.40966553,  0.59033447]])

    This method also works if the elevations are passed as an array instead of
    the (implied) field name 'topographic__elevation'.

    >>> z = grid['node']['topographic__elevation']
    >>> (receivers, proportions,
    ... steepest_slope, steepest_receiver,
    ... sink, receiver_links, steepest_link) = flow_directions_dinf(grid, z)
    >>> receivers
    array([[ 0, -1],
           [ 0,  3],
           [ 1,  4],
           [ 0,  1],
           [ 3,  0],
           [ 4,  1],
           [ 3,  4],
           [ 6,  3],
           [ 7,  4]])

    >>> proportions
    array([[ 1.        ,  0.        ],
           [ 1.        , -0.        ],
           [ 1.        , -0.        ],
           [ 1.        ,  0.        ],
           [ 0.40966553,  0.59033447],
           [ 0.40966553,  0.59033447],
           [ 1.        ,  0.        ],
           [ 0.40966553,  0.59033447],
           [ 0.40966553,  0.59033447]])
    """
    # grid type testing
    if isinstance(grid, VoronoiDelaunayGrid):
        raise NotImplementedError('Dinfinity is currently implemented for'
                                  ' Raster grids only')
    # get elevs
    elevs = return_array_at_node(grid, elevs)

    ### Step 1, some basic set-up, gathering information about the grid.

    # Calculate the number of nodes.
    num_nodes = len(elevs)

    # Set the number of receivers and facets.
    num_receivers = 2
    num_facets = 8

    # Create a node array
    node_id = np.arange(num_nodes)

    # find where there are closed nodes.
    closed_nodes = grid.status_at_node == CLOSED_BOUNDARY

    # create an array of the triangle numbers
    tri_numbers = np.arange(num_facets)

    ### Step 3, create some triangle datastructures because landlab (smartly)
    # makes it hard to deal with diagonals.

    # create list of triangle neighbors at node. Use orientation associated
    # with tarboton's 1997 algorithm, orthogonal link first, then diagonal.
    # has shape, (nnodes, 8 triangles, 2 neighbors)
    n_at_node = grid.adjacent_nodes_at_node
    dn_at_node = grid.diagonal_adjacent_nodes_at_node

    triangle_neighbors_at_node = np.stack([np.vstack((n_at_node[:,0], dn_at_node[:,0])),
                                           np.vstack((n_at_node[:,1], dn_at_node[:,0])),
                                           np.vstack((n_at_node[:,1], dn_at_node[:,1])),
                                           np.vstack((n_at_node[:,2], dn_at_node[:,1])),
                                           np.vstack((n_at_node[:,2], dn_at_node[:,2])),
                                           np.vstack((n_at_node[:,3], dn_at_node[:,2])),
                                           np.vstack((n_at_node[:,3], dn_at_node[:,3])),
                                           np.vstack((n_at_node[:,0], dn_at_node[:,3]))],
                                          axis=-1)
    triangle_neighbors_at_node = triangle_neighbors_at_node.swapaxes(0,1)

    # next create, triangle links at node
    l_at_node = grid.d8s_at_node[:, :4]
    dl_at_node = grid.d8s_at_node[:, 4:]
    triangle_links_at_node = np.stack([np.vstack((l_at_node[:,0], dl_at_node[:,0])),
                                       np.vstack((l_at_node[:,1], dl_at_node[:,0])),
                                       np.vstack((l_at_node[:,1], dl_at_node[:,1])),
                                       np.vstack((l_at_node[:,2], dl_at_node[:,1])),
                                       np.vstack((l_at_node[:,2], dl_at_node[:,2])),
                                       np.vstack((l_at_node[:,3], dl_at_node[:,2])),
                                       np.vstack((l_at_node[:,3], dl_at_node[:,3])),
                                       np.vstack((l_at_node[:,0], dl_at_node[:,3]))],
                                      axis=-1)
    triangle_links_at_node = triangle_links_at_node.swapaxes(0,1)

    # next create link directions and active link directions at node
    # link directions
    ld_at_node = grid.link_dirs_at_node
    dld_at_node = grid.diagonal_dirs_at_node
    triangle_link_dirs_at_node = np.stack([np.vstack((ld_at_node[:,0], dld_at_node[:,0])),
                                           np.vstack((ld_at_node[:,1], dld_at_node[:,0])),
                                           np.vstack((ld_at_node[:,1], dld_at_node[:,1])),
                                           np.vstack((ld_at_node[:,2], dld_at_node[:,1])),
                                           np.vstack((ld_at_node[:,2], dld_at_node[:,2])),
                                           np.vstack((ld_at_node[:,3], dld_at_node[:,2])),
                                           np.vstack((ld_at_node[:,3], dld_at_node[:,3])),
                                           np.vstack((ld_at_node[:,0], dld_at_node[:,3]))],
                                          axis=-1)
    triangle_link_dirs_at_node = triangle_link_dirs_at_node.swapaxes(0,1)

#    # active link directions.
#    ald_at_node = grid.active_link_dirs_at_node
#    adld_at_node = grid.active_diagonal_dirs_at_node
#
#    triangle_active_link_dirs_at_node = np.stack([np.vstack((ald_at_node[:,0], adld_at_node[:,0])),
#                                                  np.vstack((ald_at_node[:,1], adld_at_node[:,0])),
#                                                  np.vstack((ald_at_node[:,1], adld_at_node[:,1])),
#                                                  np.vstack((ald_at_node[:,2], adld_at_node[:,1])),
#                                                  np.vstack((ald_at_node[:,2], adld_at_node[:,2])),
#                                                  np.vstack((ald_at_node[:,3], adld_at_node[:,2])),
#                                                  np.vstack((ald_at_node[:,3], adld_at_node[:,3])),
#                                                  np.vstack((ald_at_node[:,0], adld_at_node[:,3]))],
#                                                 axis=-1)
#    triangle_active_link_dirs_at_node = triangle_active_link_dirs_at_node.swapaxes(0,1)
#
    # need to create a list of diagonal links since it doesn't exist.
    diag_links = np.sort(np.unique(grid.d8s_at_node[:, 4:]))
    diag_links = diag_links[diag_links>0]

    # calculate graidents across diagonals and orthogonals
    diag_grads = grid._calculate_gradients_at_d8_links(elevs)
    ortho_grads = grid.calc_grad_at_link(elevs)

    # finally compile link slopes
    link_slope = np.hstack((ortho_grads,
                            diag_grads))

    # Construct the array of slope to triangles at node. This also will adjust
    # for the slope convention based on the direction of the links.
    # this is a (nnodes, 2, 8) array
    slopes_to_triangles_at_node = link_slope[triangle_links_at_node]*triangle_link_dirs_at_node

    # identify where nodes are closed.
    closed_triangle_neighbors = closed_nodes[triangle_neighbors_at_node]
    # construct some arrays that deal with the distances between points on the
    # grid.

    #### Step 3: make arrays necessary for the specific tarboton algorithm.
    # create a arrays
    ac = np.array([0., 1., 1., 2., 2., 3., 3., 4.])
    af = np.array([1., -1., 1., -1., 1., -1., 1., -1.])

    # construct d1 and d2, we know these because we know where the orthogonal
    # links are
    diag_length = ((grid.dx)**2+(grid.dy)**2)**0.5

    # for irregular grids, d1 and d2 will need to be matricies
    d1 = np.array([grid.dx, grid.dy, grid.dy, grid.dx, grid.dx, grid.dy, grid.dy, grid.dy])
    d2 = np.array([grid.dx, grid.dx, grid.dy, grid.dy, grid.dx, grid.dx, grid.dy, grid.dy])

    thresh = np.arctan(d2/d1)

    ##### Step 4, Initialize receiver and proportion arrays
    receivers = UNDEFINED_INDEX * np.ones((num_nodes, num_receivers), dtype=int)
    receiver_closed = UNDEFINED_INDEX * np.ones((num_nodes, num_receivers), dtype=int)
    proportions = np.zeros((num_nodes, num_receivers), dtype=float)
    receiver_links = UNDEFINED_INDEX * np.ones((num_nodes, num_receivers), dtype=int)
    slopes_to_receivers = np.zeros((num_nodes, num_receivers), dtype=float)

    #### Step  5  begin the algorithm in earnest

    # construct e0, e1, e2 for all triangles at all nodes.
    # will be (nnodes, nfacets=8 for raster or nfacets = max number of patches
    # for irregular grids.

    # e0 is origin point of the facet
    e0 = elevs[node_id]

    # e1 is the point on the orthogoanal edges
    e1 = elevs[triangle_neighbors_at_node[:,0,:]]
    # e2 is the point on the diagonal edges
    e2 = elevs[triangle_neighbors_at_node[:,1,:]]

    # mask out where nodes do not exits (e.g. triangle_neighbors_at_node == -1)
    e2[triangle_neighbors_at_node[:,1,:] == -1] = np.nan
    e1[triangle_neighbors_at_node[:,0,:] == -1] = np.nan

    # loop through and calculate s1 and s2
    # this will only loop nfacets times.
    s1 = np.empty_like(e1)
    s2 = np.empty_like(e2)

    for i in range(num_facets):
        s1[:,i] = (e0 - e1[:, i])/d1[i]
        s2[:,i] = (e1[:, i]- e2[:, i])/d2[i]

    # calculate r and s, the direction and magnitude
    r = np.arctan2(s2, s1)
    s = ((s1**2)+(s2**2))**0.5

    r[np.isnan(r)] = 0
    # adjust r if it doesn't sit in the realm of (0, arctan(d2,d1))
    too_small = r < 0
    radj = r.copy()
    radj[too_small] = 0
    s[too_small] = s1[too_small]

    # to consider two big, we need to look by trangle.
    for i in range(num_facets):
        too_big = r[:, i] > thresh[i]
        radj[too_big, i] = thresh[i]
        s[too_big, i] = (e0[too_big] - e2[too_big, i])/diag_length

    # calculate the geospatial version of r based on radj
    rg = np.empty_like(r)
    for i in range(num_facets):
        rg[:, i] = (af[i]*radj[:, i]) + (ac[i]*np.pi/2.)

    # set slopes that are nan to zero
    s[np.isnan(s)] = 0

    # sort slopes based on
    steepest_sort = np.argsort(s)

    # determine the steepest triangle
    steepest_triangle = tri_numbers[steepest_sort[:, -1]]

    # initialize arrays for the steepest rg and steepest s
    steepest_rg = np.empty_like(node_id, dtype = float)
    steepest_s = np.empty_like(node_id, dtype = float)

    for n in node_id:
        steepest_rg[n] = rg[n, steepest_sort[n, -1]]
        receiver_closed[n] = closed_triangle_neighbors[n, :, steepest_sort[n, -1]]
        steepest_s[n] = s[n, steepest_sort[n, -1]]
        receivers[n, :] = triangle_neighbors_at_node[n, :, steepest_sort[n, -1]]
        receiver_links[n, :] = triangle_links_at_node[n, :, steepest_sort[n, -1]]
        slopes_to_receivers[n, :] = slopes_to_triangles_at_node[n, :, steepest_sort[n, -1]]

    # construct the baseline for proportions
    rg_baseline = np.array([0., 1., 1., 2., 2., 3., 3., 4])*np.pi/2.
    #rg_baseline = np.array([0., 0.5, 1., 1.5, 2., 2.5, 3., 3.5])*np.pi/4.

    # calculate alpha1 and alpha 2
    alpha2 = (steepest_rg-rg_baseline[steepest_triangle])*af[steepest_triangle]
    alpha1 = thresh[steepest_triangle] - alpha2

    # calculate proportions from alpha
    proportions[:, 0] = (alpha1)/(alpha1+alpha2)
    proportions[:, 1] = (alpha2)/(alpha1+alpha2)

    ### END OF THE Tarboton algorithm, start of work to make this code mesh
    # with other landlab flow directing algorithms.

    # identify what drains to itself, and set proportion and id values based on
    # that.

    # if proportions is nan, drain to self
    drains_to_self = np.isnan(proportions[:, 0])

    # if all slopes are leading out, drain to self
    drains_to_self[steepest_s <= 0] = True

    # if both receiver nodes are closed, drain to self
    drains_to_two_closed = receiver_closed.sum(axis=1) == num_receivers
    drains_to_self[drains_to_two_closed] = True

    # if drains to one closed receiver, check that the open receiver actually
    # gets flow. If so, route all to the open receiver. If the receiver getting
    # all the flow is closed, then drain to self.
    all_flow_to_closed = np.sum(receiver_closed*proportions, axis=1) == 1
    drains_to_self[all_flow_to_closed] = True

    drains_to_one_closed = receiver_closed.sum(axis=1)==1
    fix_flow = drains_to_one_closed * (all_flow_to_closed == False)
    first_column_has_closed = np.array(receiver_closed[:, 0]*fix_flow,
                                       dtype=bool)
    second_column_has_closed = np.array(receiver_closed[:, 1]*fix_flow,
                                        dtype=bool)

    # remove the link to the closed node
    receivers[first_column_has_closed, 0] = -1
    receivers[second_column_has_closed, 1] = -1

    # change the proportions
    proportions[first_column_has_closed,0] = 0.
    proportions[first_column_has_closed,1] = 1.

    proportions[second_column_has_closed, 0] = 1.
    proportions[second_column_has_closed, 1] = 0.

    # set properties of drains to self.
    receivers[drains_to_self, 0] = node_id[drains_to_self]
    receivers[drains_to_self, 1] = -1

    proportions[drains_to_self, 0] = 1.
    proportions[drains_to_self, 1] = 0.

    # mask the receiver_links by where flow doesn't occur to return
    receiver_links[drains_to_self, :] = UNDEFINED_INDEX

    # identify the steepest link so that the steepest receiver, link, and slope
    # can be returned.
    slope_sort = np.argsort(np.argsort(slopes_to_receivers,
                                       axis=1),
                            axis=1) == (num_receivers-1)
    steepest_slope = slopes_to_receivers[slope_sort]
    steepest_slope[drains_to_self] = 0.

    ## identify the steepest link and steepest receiever.
    steepest_link = receiver_links[slope_sort]
    steepest_link[drains_to_self] = UNDEFINED_INDEX

    steepest_receiver = receivers[slope_sort]
    steepest_receiver[drains_to_self] = node_id[drains_to_self]


    # Optionally, handle baselevel nodes: they are their own receivers
    if baselevel_nodes is not None:
        receivers[baselevel_nodes,0] = node_id[baselevel_nodes]
        receivers[baselevel_nodes,1:] = -1
        proportions[baselevel_nodes, 0] = 1
        proportions[baselevel_nodes, 1:] = 0
        receiver_links[baselevel_nodes,:] = UNDEFINED_INDEX
        steepest_slope[baselevel_nodes] = 0.

    # The sink nodes are those that are their own receivers (this will normally
    # include boundary nodes as well as interior ones; "pits" would be sink
    # nodes that are also interior nodes).
    (sink, ) = np.where(node_id == receivers[:, 0])
    sink = as_id_array(sink)

    return (receivers, proportions, steepest_slope, steepest_receiver, sink,
            receiver_links, steepest_link)

if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
This is an implementation of Vaughan Voller's experimental boundary method
reduced complexity flow router. Credit: Voller, Hobley, Paola.

Created on Fri Feb 20 09:32:27 2015

@author: danhobley (SiccarPoint), after volle001@umn.edu
"""

# ##in the diagonal case, even closed edges can produce "drag". Is this right?
# Could suppress by mirroring the diagonals

import numpy as np
from landlab import (RasterModelGrid, Component, FieldError, INACTIVE_LINK,
                     ACTIVE_LINK, CLOSED_BOUNDARY, CORE_NODE, FIXED_LINK)
import inspect
from landlab.utils.decorators import use_file_name_or_kwds


class PotentialityFlowRouter(Component):
    """Multidirectional flow routing using a novel method.

    This class implements Voller, Hobley, and Paola's experimental matrix
    solutions for flow routing. The method works by solving for a potential
    field at all nodes on the grid, which enforces both mass conservation
    and flow downhill along topographic gradients. It is order n and highly
    efficient, but does not return any information about flow connectivity.

    Options are permitted to allow "abstract" routing (flow enforced downslope,
    but no particular assumptions are made about the governing equations), or
    routing according to the Chezy or Manning equations. This routine assumes
    that water is distributed evenly over the surface of the cell in deriving
    the depth, and does not assume channelization. You will need to back-
    calculate channel depths for yourself using known widths at each node
    if that is what you want.

    It is VITAL you initialize this component AFTER setting boundary
    conditions.

    Note that this component offers the property `discharges_at_links`. This
    returns the discharges at all links. If method=='D8', this list will
    include diagonal links after the orthogonal links, which is why this
    information is not returned as a field.

    Discharges at nodes are recorded as the outgoing total discharge (i.e.,
    including any contribution from 'water__unit_flux_in').

    The primary method of this class is :func:`run_one_step`.

    Notes
    -----
    This is a "research grade" component, and is subject to dramatic change
    with little warning. No guarantees are made regarding its accuracy or
    utility. It is not recommended for user use yet!

    Examples
    --------
    >>> from landlab import HexModelGrid
    >>> import numpy as np
    >>> mg = HexModelGrid(4, 6, dx=2., shape='rect', orientation='vertical')
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> Q_in = mg.add_ones('node', 'water__unit_flux_in')
    >>> z += mg.node_y.copy()
    >>> potfr = PotentialityFlowRouter(mg)
    >>> potfr.run_one_step()
    >>> Q_at_core_nodes = np.array(
    ...     [ 13.57233404,  13.93522481,  11.52216193,  11.29307277,
    ...        8.80884751,   8.86380667,   6.47446459,   6.82161521])
    >>> np.allclose(mg.at_node['surface_water__discharge'][mg.core_nodes],
    ...             Q_at_core_nodes)
    True
    """
#    >>> Q_at_core_nodes = np.array(
#    ...     [ 17.02012846,  16.88791903,  13.65746194,  14.85578934,
#    ...       11.41908145,  11.43630865,   8.95902559,  10.04348075,
#    ...        6.28696459,   6.44316089,   4.62478522,   5.29145188])
    _name = 'PotentialityFlowRouter'

    _input_var_names = ('topographic__elevation',
                        'water__unit_flux_in',
                        )

    _output_var_names = ('surface_water__discharge',
                         'flow__potential',
                         'surface_water__depth',
                         )

    _var_units = {'topographic__elevation': 'm',
                  'water__unit_flux_in': 'm/s',
                  'surface_water__discharge': 'm**3/s',
                  'flow__potential': 'm**3/s',
                  'surface_water__depth': 'm',
                  }

    _var_mapping = {'topographic__elevation': 'node',
                    'water__unit_flux_in': 'node',
                    'surface_water__discharge': 'node',
                    'flow__potential': 'node',
                    'surface_water__depth': 'node',
                    }

    _var_doc = {
        'topographic__elevation': 'Land surface topographic elevation',
        'water__unit_flux_in': (
            'External volume water per area per time input to each node ' +
            '(e.g., rainfall rate)'),
        'surface_water__discharge': (
            'Magnitude of volumetric water flux out of each node'),
        'flow__potential': (
            'Value of the hypothetical field "K", used to force water flux ' +
            'to flow downhill'),
        'surface_water__depth': (
            'If Manning or Chezy specified, the depth of flow in the cell, ' +
            'calculated assuming flow occurs over the whole surface'),
                  }

    _min_slope_thresh = 1.e-24
    # if your flow isn't connecting up, this probably needs to be reduced

    @use_file_name_or_kwds
    def __init__(self, grid, method='D8', flow_equation='default',
                 Chezys_C=30., Mannings_n=0.03, **kwds):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        method : {'D8', 'D4'}, optional
            Routing method ('D8' is the default). This keyword has no effect for a
            Voronoi-based grid.
        flow_equation : {'default', 'Manning', 'Chezy'}, optional
            If Manning or Chezy, flow is routed according to the Manning or Chezy
            equation; discharge is allocated to multiple downslope nodes
            proportional to the square root of discharge; and a water__depth field
            is returned. If default, flow is allocated to multiple nodes linearly
            with slope; and the water__depth field is not calculated.
        Chezys_C : float (optional)
            Required if flow_equation == 'Chezy'.
        Mannings_n : float (optional)
            Required if flow_equation == 'Manning'.
        """
        if RasterModelGrid in inspect.getmro(grid.__class__):
            assert grid.number_of_node_rows >= 3
            assert grid.number_of_node_columns >= 3
            self._raster = True
        else:
            self._raster = False

        self._grid = grid
        self.equation = flow_equation
        assert self.equation in ('default', 'Chezy', 'Manning')
        if self.equation == 'Chezy':
            self.chezy_C = Chezys_C
        elif self.equation == 'Manning':
            self.manning_n = Mannings_n
        assert method in ('D8', 'D4')
        if method == 'D8':
            self.route_on_diagonals = True
        else:
            self.route_on_diagonals = False

        # hacky fix because water__discharge is defined on both links and nodes
        for out_field in self._output_var_names:
            if self._var_mapping[out_field] == 'node':
                try:
                    self.grid.add_zeros(self._var_mapping[out_field],
                                        out_field, dtype=float)
                except FieldError:
                    pass
            else:
                pass
            try:
                self.grid.add_zeros('node', 'surface_water__discharge', dtype=float)
            except FieldError:
                pass

        if self._raster:
            self.equiv_circ_diam = 2. * np.sqrt(grid.dx * grid.dy / np.pi)
        else:
            for_cell_areas = 2. * np.sqrt(grid.area_of_cell / np.pi)
            mean_A = for_cell_areas.mean()
            self.equiv_circ_diam = for_cell_areas[grid.cell_at_node]
            self.equiv_circ_diam[grid.cell_at_node == -1] = mean_A
        # ^this is the equivalent seen CSWidth of a cell for a flow in a
        # generic 360 direction
        if self.route_on_diagonals and self._raster:
            self._discharges_at_link = np.empty(grid.number_of_d8)
        else:
            self._discharges_at_link = self.grid.empty('link')

    def route_flow(self, **kwds):
        """
        """
        grid = self.grid
        self._K = grid.at_node['flow__potential']
        self._Qw = grid.at_node['surface_water__discharge']
        z = grid.at_node['topographic__elevation']
        qwater_in = grid.at_node['water__unit_flux_in'].copy()
        qwater_in[grid.node_at_cell] *= grid.area_of_cell
        prev_K = self._K.copy()
        mismatch = 10000.
        # do the ortho nodes first, in isolation
        g = grid.calc_grad_at_link(z)
        if self.equation != 'default':
            g = np.sign(g) * np.sqrt(np.fabs(g))
            # ^...because both Manning and Chezy actually follow sqrt
            # slope, not slope
        # weight by face width - NO, because diags
        # g *= grid.width_of_face[grid.face_at_link]
        link_grad_at_node_w_dir = (
            g[grid.links_at_node] * grid.active_link_dirs_at_node)
        # active_link_dirs strips "wrong" face widths

        # now outgoing link grad sum
        outgoing_sum = (np.sum((link_grad_at_node_w_dir).clip(0.), axis=1) +
                        self._min_slope_thresh)
        pos_incoming_link_grads = (-link_grad_at_node_w_dir).clip(0.)

        if not self.route_on_diagonals or not self._raster:
            while mismatch > 1.e-6:
                K_link_ends = self._K[grid.adjacent_nodes_at_node]
                incoming_K_sum = (pos_incoming_link_grads*K_link_ends
                                  ).sum(axis=1) + self._min_slope_thresh
                self._K[:] = (incoming_K_sum + qwater_in)/outgoing_sum
                mismatch = np.sum(np.square(self._K-prev_K))
                prev_K = self._K.copy()

            upwind_K = grid.map_value_at_max_node_to_link(z, self._K)
            self._discharges_at_link[:] = upwind_K * g
            self._discharges_at_link[grid.status_at_link == INACTIVE_LINK] = 0.
        else:
            # grad on diags:
            gwd = np.empty(grid.number_of_d8, dtype=float)
            gd = gwd[grid.number_of_links:]
            gd[:] = (z[grid.nodes_at_diagonal[:, 1]] - z[grid.nodes_at_diagonal[:, 0]])
            gd /= grid.length_of_d8[grid.number_of_links:]
            if self.equation != 'default':
                gd[:] = np.sign(gd)*np.sqrt(np.fabs(gd))
            diag_grad_at_node_w_dir = (gwd[grid.d8s_at_node[:, 4:]] *
                                       self.grid.active_diagonal_dirs_at_node)

            outgoing_sum += np.sum(diag_grad_at_node_w_dir.clip(0.), axis=1)
            pos_incoming_diag_grads = (-diag_grad_at_node_w_dir).clip(0.)
            while mismatch > 1.e-6:
                K_link_ends = self._K[grid.adjacent_nodes_at_node]
                K_diag_ends = self._K[grid.diagonal_adjacent_nodes_at_node]
                incoming_K_sum = ((pos_incoming_link_grads * K_link_ends
                                   ).sum(axis=1) +
                                  (pos_incoming_diag_grads * K_diag_ends
                                   ).sum(axis=1) + self._min_slope_thresh)
                self._K[:] = (incoming_K_sum + qwater_in) / outgoing_sum
                mismatch = np.sum(np.square(self._K - prev_K))
                prev_K = self._K.copy()

            # ^this is necessary to suppress stupid apparent link Qs at flow
            # edges, if present.
            upwind_K = grid.map_value_at_max_node_to_link(z, self._K)
            upwind_diag_K = np.where(
                z[grid.nodes_at_diagonal[:, 1]] > z[grid.nodes_at_diagonal[:, 0]],
                self._K[grid.nodes_at_diagonal[:, 1]],
                self._K[grid.nodes_at_diagonal[:, 0]])
            self._discharges_at_link[:grid.number_of_links] = upwind_K * g
            self._discharges_at_link[grid.number_of_links:] = (
                upwind_diag_K * gd)
            self._discharges_at_link[grid.status_at_d8 == FIXED_LINK] = 0.

        np.multiply(self._K, outgoing_sum, out=self._Qw)
        # there is no sensible way to save discharges at links, if we route
        # on diagonals.
        # for now, let's make a property

        # now process uval and vval to give the depths, if Chezy or Manning:
        if self.equation == 'Chezy':
            # Chezy: Q = C*Area*sqrt(depth*slope)
            grid.at_node['surface_water__depth'][:] = (
                grid.at_node['flow__potential'] / self.chezy_C /
                self.equiv_circ_diam) ** (2. / 3.)
        elif self.equation == 'Manning':
            # Manning: Q = w/n*depth**(5/3)
            grid.at_node['surface_water__depth'][:] = (
                grid.at_node['flow__potential'] * self.manning_n /
                self.equiv_circ_diam) ** 0.6
        else:
            pass

    def run_one_step(self, **kwds):
        """Route surface-water flow over a landscape.

        Both convergent and divergent flow can occur.
        """
        self.route_flow(**kwds)

    @property
    def discharges_at_links(self):
        """Return the discharges at links.

        Note that if diagonal routing, this will return number_of_d8.
        Otherwise, it will be number_of_links.
        """
        return self._discharges_at_link



# -*- coding: utf-8 -*-
"""
Landlab component for overland flow using a local implicit solution to the
kinematic-wave approximation.

Created on Fri May 27 14:26:13 2016

@author: gtucker
"""


from landlab import Component
from landlab.components import FlowAccumulator
from scipy.optimize import newton
import numpy as np


def water_fn(x, a, b, c, d, e):
    """Evaluates the solution to the water-depth equation.

    Called by scipy.newton() to find solution for :math:`x`
    using Newton's method.

    Parameters
    ----------
    x : float
        Water depth at new time step.
    a : float
        "alpha" parameter (see below)
    b : float
        Weighting factor on new versus old time step. :math:`b=1` means purely
        implicit solution with all weight on :math:`H` at new time
        step. :math:`b=0` (not recommended) would mean purely explicit.
    c : float
        Water depth at old time step (time step :math:`t` instead
        of :math:`t+1`)
    d : float
        Depth-discharge exponent; normally either 5/3 (Manning) or 3/2 (Chezy)
    e : float
        Water inflow volume per unit cell area in one time step.

    This equation represents the implicit solution for water depth
    :math:`H` at the next time step. In the code below, it is
    formulated in a generic way.  Written using more familiar
    terminology, the equation is:

    .. math::

        H - H_0 + \alpha ( w H + (w-1) H_0)^d - \Delta t (R + Q_{in} / A)

    .. math::

        \alpha = \frac{\Delta t \sum S^{1/2}}{C_f A}
    
    where :math:`H` is water depth at the given node at the new
    time step, :math:`H_0` is water depth at the prior time step,
    :math:`w` is a weighting factor, :math:`d` is the depth-discharge
    exponent (2/3 or 1/2), :math:`\Delta t` is time-step duration,
    :math:`R` is local runoff rate, :math:`Q_{in}` is inflow
    discharge, :math:`A` is cell area, :math:`C_f` is a
    dimensional roughness coefficient, and :math:`\sum S^{1/2}`
    represents the sum of square-root-of-downhill-gradient over
    all outgoing (downhill) links.
    """
    return x - c + a * (b * x + (b - 1.0) * c) ** d - e


class KinwaveImplicitOverlandFlow(Component):
    """Calculate shallow water flow over topography.

    Landlab component that implements a two-dimensional kinematic wave model.
    This is a form of the 2D shallow-water equations in which energy slope is
    assumed to equal bed slope. The solution method is locally implicit, and
    works as follows. At each time step, we iterate from upstream to downstream
    over the topography. Because we are working downstream, we can assume that
    we know the total water inflow to a given cell. We solve the following mass 
    conservation equation at each cell:
    
    .. math::

        (H^{t+1} - H^t)/\Delta t = Q_{in}/A - Q_{out}/A + R
        
    where :math:`H` is water depth, :math:`t` indicates time step
    number, :math:`\Delta t` is time step duration, :math:`Q_{in}` is
    total inflow discharge, :math:`Q_{out}` is total outflow
    discharge, :math:`A` is cell area, and :math:`R` is local
    runoff rate (precipitation minus infiltration; could be
    negative if runon infiltration is occurring).
    
    The specific outflow discharge leaving a cell along one of its faces is:

    .. math::

        q = (1/C_r) H^\alpha S^{1/2}
    
    where :math:`C_r` is a roughness coefficient (such as
    Manning's n), :math:`\alpha` is an exponent equal to :math:`5/3`
    for the Manning equation and :math:`3/2` for the Chezy family,
    and :math:`S` is the downhill-positive gradient of the link
    that crosses this particular face. Outflow discharge is zero
    for links that are flat or "uphill" from the given node.
    Total discharge out of a cell is then the sum of (specific
    discharge x face width) over all outflow faces

    .. math::

        Q_{out} = \sum_{i=1}^N (1/C_r) H^\alpha S_i^{1/2} W_i
        
    where :math:`N` is the number of outflow faces (i.e., faces
    where the ground slopes downhill away from the cell's node),
    and :math:`W_i` is the width of face :math:`i`.
    
    We use the depth at the cell's node, so this simplifies to:

    .. math::

        Q_{out} = (1/C_r) H'^\alpha \sum_{i=1}^N S_i^{1/2} W_i
        
    We define :math:`H` in the above as a weighted sum of
    the "old" (time step :math:`t`) and "new" (time step :math:`t+1`)
    depth values:

    .. math::

        H' = w H^{t+1} + (1-w) H^t
    
    If :math:`w=1`, the method is fully implicit. If :math:`w=0`,
    it is a simple forward explicit method.
    
    When we combine these equations, we have an equation that includes the
    unknown :math:`H^{t+1}` and a bunch of terms that are known.
    If :math:`w\ne 0`, it is a nonlinear equation in :math:`H^{t+1}`,
    and must be solved iteratively. We do this using a root-finding
    method in the scipy.optimize library.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> rg = RasterModelGrid((4, 5), 10.0)
    >>> z = rg.add_zeros('node', 'topographic__elevation')
    >>> kw = KinwaveImplicitOverlandFlow(rg)
    >>> round(kw.runoff_rate * 1.0e7, 2)
    2.78
    >>> kw.vel_coef  # default value
    100.0
    >>> rg.at_node['surface_water__depth'][6:9]
    array([ 0.,  0.,  0.])
    """

    _name = 'KinwaveImplicitOverlandFlow'

    _input_var_names = (
        'topographic__elevation',
    )

    _output_var_names = (
        'topographic__gradient',
        'surface_water__depth',
        #'water__velocity',
        #'water__specific_discharge',
        'surface_water_inflow__discharge',
    )

    _var_units = {
        'topographic__elevation': 'm',
        'topographic__slope': 'm/m',
        'surface_water__depth': 'm',
        'water__velocity': 'm/s',
        'water__specific_discharge': 'm2/s',
    }

    _var_mapping = {
        'topographic__elevation': 'node',
        'topographic__gradient': 'link',
        'surface_water__depth': 'node',
        #'water__velocity': 'link',
        #'water__specific_discharge': 'link',
        'surface_water_inflow__discharge' : 'node',
    }

    _var_doc = {
        'topographic__elevation':
            'elevation of the ground surface relative to some datum',
        'topographic__gradient':
            'gradient of the ground surface',
        'surface_water__depth':
            'depth of water',
#        'water__velocity':
#            'flow velocity component in the direction of the link',
#        'water__specific_discharge':
#            'flow discharge component in the direction of the link',
        'surface_water_inflow__discharge':
            'water volume inflow rate to the cell around each node'
    }

    def __init__(self, grid, runoff_rate=1.0, roughness=0.01,
                 changing_topo=False, depth_exp=1.5, weight=1.0, **kwds):
        """Initialize the KinwaveOverlandFlowModel.

        Parameters
        ----------
        grid : ModelGrid
            Landlab ModelGrid object
        runoff_rate : float, optional (defaults to 1 mm/hr)
            Precipitation rate, mm/hr
        roughnes : float, defaults to 0.01
            Manning roughness coefficient, s/m^1/3
        changing_topo : boolean, optional (defaults to False)
            Flag indicating whether topography changes between time steps
        depth_exp : float (defaults to 1.5)
            Exponent on water depth in velocity equation (3/2 for Darcy/Chezy,
            5/3 for Manning)
        weight : float (defaults to 1.0)
            Weighting on depth at new time step versus old time step (1 = all
            implicit; 0 = explicit)
        """

        # Store grid and parameters and do unit conversion
        self._grid = grid
        self.runoff_rate = runoff_rate / 3600000.0  # convert to m/s
        self.vel_coef = 1.0 / roughness  # do division now to save time
        self.changing_topo = changing_topo
        self.depth_exp = depth_exp
        self.weight = weight

        # Get elevation field
        try:
            self.elev = grid.at_node['topographic__elevation']
        except:
            raise

        # Create fields...
        #  Water depth
        if 'surface_water__depth' in grid.at_node:
            self.depth = grid.at_node['surface_water__depth']
        else:
            self.depth = grid.add_zeros('node', 'surface_water__depth')
        #   Slope
        if 'topographic__gradient' in grid.at_link:
            self.slope = grid.at_link['topographic__gradient']
        else:
            self.slope = grid.add_zeros('link', 'topographic__gradient')
        #  Velocity
#        if 'water__velocity' in grid.at_link:
#            self.vel = grid.at_link['water__velocity']
#        else:
#            self.vel = grid.add_zeros('link', 'water__velocity')
        #  Discharge
#        if 'surface_water__specific_discharge' in grid.at_link:
#            self.disch = grid.at_link['surface_water__specific_discharge']
#        else:
#            self.disch = grid.add_zeros('link',
#                                        'surface_water__specific_discharge')
        #  Inflow discharge at nodes
        if 'surface_water_inflow__discharge' in grid.at_node:
            self.disch_in = grid.at_node['surface_water_inflow__discharge']
        else:
            self.disch_in = grid.add_zeros('node',
                                           'surface_water_inflow__discharge')

        # This array holds, for each node, the sum of sqrt(slope) x face width
        # for each link/face.
        self.grad_width_sum = grid.zeros('node')

        # This array holds the prefactor in the algebraic equation that we
        # will find a solution for.
        self.alpha = grid.zeros('node')

        # Instantiate flow router
        self.flow_accum = FlowAccumulator(grid, 'topographic__elevation',
                                    flow_director='MFD',
                                    partition_method='square_root_of_slope')

        # Flag to let us know whether this is our first iteration
        self.first_iteration = True

    def run_one_step(self, dt, current_time=0.0, runoff_rate=None, **kwds):
        """Calculate water flow for a time period `dt`.
        """
        # Handle runoff rate
        if runoff_rate is None:
            runoff_rate = self.runoff_rate

        # If it's our first iteration, or if the topography may be changing,
        # do flow routing and calculate square root of slopes at links     
        if self.changing_topo or self.first_iteration:

            # Calculate the ground-surface slope
            self.slope[self.grid.active_links] = \
                self._grid.calc_grad_at_link(self.elev)[self._grid.active_links]
                
            # Take square root of slope magnitude for use in velocity eqn
            self.sqrt_slope = np.sqrt(np.abs(self.slope))

            # Re-route flow, which gives us the downstream-to-upstream
            # ordering
            self.flow_accum.run_one_step()
            self.nodes_ordered = self.grid.at_node['flow__upstream_node_order']
            self.flow_lnks = self.grid.at_node['flow__links_to_receiver_nodes']

            # (Re)calculate, for each node, sum of sqrt(gradient) x width
            self.grad_width_sum[:] = 0.0
            for i in range(self.flow_lnks.shape[1]):
                self.grad_width_sum[:] += (self.sqrt_slope[self.flow_lnks[:,i]]
                    * self._grid.width_of_face[
                            self.grid.face_at_link[self.flow_lnks[:,i]]])

            # Calculate values of alpha, which is defined as
            #
            #   $\alpha = \frac{\Sigma W S^{1/2} \Delta t}{A C_r}$
            cores = self.grid.core_nodes
            self.alpha[cores] = (
                self.vel_coef * self.grad_width_sum[cores] * dt 
                / (self.grid.area_of_cell[self.grid.cell_at_node[cores]]))

        # Zero out inflow discharge
        self.disch_in[:] = 0.0

        # Upstream-to-downstream loop
        for i in range(len(self.nodes_ordered) - 1, -1, -1):
            n = self.nodes_ordered[i]
            if self.grid.status_at_node[n] == 0:

                # Solve for new water depth
                aa = self.alpha[n]
                cc = self.depth[n]
                ee = ((dt * runoff_rate) 
                       + (dt * self.disch_in[n] 
                          / self.grid.area_of_cell[self.grid.cell_at_node[n]]))
                self.depth[n] = newton(water_fn, self.depth[n], 
                                       args=(aa, self.weight, cc, 
                                             self.depth_exp, ee))

                # Calc outflow
                Heff = (self.weight * self.depth[n]
                        + (1.0 - self.weight) * cc)
                outflow = (self.vel_coef * (Heff ** self.depth_exp)
                           * self.grad_width_sum[n])  # this is manning/chezy/darcy

                # Send flow downstream. Here we take total inflow discharge
                # and partition it among the node's neighbors. For this, we use
                # the flow director's "proportions" array, which contains, for
                # each node, the proportion of flow that heads out toward each
                # of its N neighbors. The proportion is zero if the neighbor is
                # uphill; otherwise, it is S^1/2 / sum(S^1/2). If for example
                # we have a raster grid, there will be four neighbors and four
                # proportions, some of which may be zero and some between 0 and
                # 1.
                self.disch_in[self.grid.adjacent_nodes_at_node[n]] += (outflow
                    * self.flow_accum.flow_director.proportions[n])

                # TODO: the above is enough to implement the solution for flow
                # depth, but it does not provide any information about flow
                # velocity or discharge on links. This could be added as an
                # optional method, perhaps done just before output.


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#! /usr/bin/env python
"""The Landlab

:Package name: TheLandlab
:Release date: 2013-03-24
:Authors: Greg Tucker, Nicole Gasparini, Erkan Istanbulluoglu, Daniel Hobley,
    Sai Nudurupati, Jordan Adams, Eric Hutton
:URL: http://csdms.colorado.edu/trac/landlab
:License: MIT
"""

from __future__ import absolute_import
import os

from numpy import set_printoptions
try:
    set_printoptions(legacy='1.13')
except TypeError:
    pass
finally:
    del set_printoptions

from ._registry import registry

cite_as = registry.format_citations

__all__ = ['registry']

if 'DISPLAY' not in os.environ:
    try:
        import matplotlib
    except ImportError:
        import warnings
        warnings.warn('matplotlib not found', ImportWarning)
    else:
        matplotlib.use('Agg')

from .core.model_parameter_dictionary import ModelParameterDictionary
from .core.model_parameter_dictionary import (MissingKeyError,
                                              ParameterValueError)
from .core.model_parameter_loader import load_params
from .core.model_component import Component
from .framework.collections import Palette, Arena, NoProvidersError
from .framework.decorators import Implements, ImplementsOrRaise
from .framework.framework import Framework
from .field.scalar_data_fields import FieldError
from .grid import *
from .plot import *

__all__.extend(['ModelParameterDictionary', 'MissingKeyError',
                'ParameterValueError', 'Component', 'Palette', 'Arena',
                'NoProvidersError', 'Implements', 'ImplementsOrRaise',
                'Framework', 'FieldError', 'LandlabTester', 'load_params'])

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid


ELEMENTS = ['node', 'cell', 'link', 'face']
# ELEMENTS += ['core_node', 'core_cell', 'active_link', 'active_face']
TYPES = ['float', 'int', 'bool']


def generate_zeros_tests():
    for element in ELEMENTS:
        for type in TYPES:
            def _test():
                rmg = RasterModelGrid(4, 5)
                number_of_elements = rmg.number_of_elements(element)
                assert_array_equal(rmg.zeros(centering=element),
                                   np.zeros(number_of_elements, dtype=np.float))
            _test.description = '%s.test_zeros_%s_%s' % (
                __name__, type, element)
            yield _test


def generate_add_zeros_tests():
    for element in ELEMENTS:
        for type in TYPES:
            def _test():
                rmg = RasterModelGrid(4, 5)
                number_of_elements = rmg.number_of_elements(element)
                rtn_values = rmg.add_zeros(element, 'name')
                assert rtn_values is rmg.field_values(element, 'name')
                assert_array_equal(
                    rtn_values, np.zeros(number_of_elements, dtype=np.float))
            _test.description = '%s.test_add_zeros_%s_%s' % (
                __name__, type, element)
            yield _test


def generate_ones_tests():
    for element in ELEMENTS:
        for type in TYPES:
            def _test():
                rmg = RasterModelGrid(4, 5)
                number_of_elements = rmg.number_of_elements(element)
                assert_array_equal(rmg.ones(centering=element),
                                   np.ones(number_of_elements, dtype=np.float))
            _test.description = '%s.test_zeros_%s_%s' % (
                __name__, type, element)
            yield _test


def generate_add_ones_tests():
    for element in ELEMENTS:
        for type in TYPES:
            def _test():
                rmg = RasterModelGrid(4, 5)
                number_of_elements = rmg.number_of_elements(element)
                rtn_values = rmg.add_ones(element, 'name')
                assert rtn_values is rmg.field_values(element, 'name')
                assert_array_equal(
                    rtn_values, np.ones(number_of_elements, dtype=np.float))
            _test.description = '%s.test_add_zeros_%s_%s' % (
                __name__, type, element)
            yield _test


def generate_empty_tests():
    elements = ['node', 'cell', 'link', 'face']
    # elements += ['core_node', 'core_cell', 'active_link', 'active_face']

    types = ['float', 'int', 'bool']

    for element in ELEMENTS:
        for type in TYPES:
            def _test():
                rmg = RasterModelGrid(4, 5)
                number_of_elements = rmg.number_of_elements(element)
                assert rmg.empty(centering=element).size == number_of_elements
            _test.description = '%s.test_zeros_%s_%s' % (
                __name__, type, element)
            yield _test


def generate_add_empty_tests():
    elements = ['node', 'cell', 'link', 'face']
    # elements += ['core_node', 'core_cell', 'active_link', 'active_face']

    types = ['float', 'int', 'bool']

    for element in ELEMENTS:
        for type in TYPES:
            def _test():
                rmg = RasterModelGrid(4, 5)
                number_of_elements = rmg.number_of_elements(element)
                rtn_values = rmg.add_empty(element, 'name')
                assert rtn_values is rmg.field_values(element, 'name')
                assert_array_equal(rtn_values.size, number_of_elements)
            _test.description = '%s.test_zeros_%s_%s' % (
                __name__, type, element)
            yield _test



#!/usr/bin/env python
"""Constants used with the netcdf module."""


_DIMENSION_NAMES = ['ni', 'nj', 'nk']
_AXES_NAMES = ['x', 'y', 'z']

_NP_TO_NC_TYPE = {
    'float32': 'f4',
    'float64': 'f8',
    'int8': 'i1',
    'int16': 'i2',
    'int32': 'i4',
    'int64': 'i8',
    'uint8': 'u1',
    'uint16': 'u2',
    'uint32': 'u4',
    'uint64': 'u8',
    'bool': 'i1',
}


_AXIS_DIMENSION_NAMES = ['nk', 'nj', 'ni']
_AXIS_COORDINATE_NAMES = ['z', 'y', 'x']

_DIMENSION_NAMES = set(_AXIS_DIMENSION_NAMES + ['nt'])
_COORDINATE_NAMES = set(_AXIS_COORDINATE_NAMES + ['t'])



from numpy.testing import assert_array_equal


class NumpyArrayTestingMixIn(object):
    def assertArrayEqual(self, actual, expected):
        try:
            assert_array_equal(actual, expected)
        except AssertionError as error:
            self.fail(error)



#!/usr/bin/env python

from landlab import RasterModelGrid
from landlab.components import FlowRouter
from landlab.utils import (get_watershed_nodes, get_watershed_outlet,
                           get_watershed_masks_with_area_threshold)
import numpy as np


def test_get_watershed_nodes():
    grid = RasterModelGrid((7, 7), 1)

    z = np.array([
        -9999., -9999., -9999., -9999., -9999., -9999., -9999.,
        -9999.,    26.,     0.,    30.,    32.,    34., -9999.,
        -9999.,    28.,     1.,    25.,    28.,    32., -9999.,
        -9999.,    30.,     3.,     3.,    11.,    34., -9999.,
        -9999.,    32.,    11.,    25.,    18.,    38., -9999.,
        -9999.,    34.,    32.,    34.,    36.,    40., -9999.,
        -9999., -9999., -9999., -9999., -9999., -9999., -9999.])

    grid.at_node['topographic__elevation'] = z

    outlet_id = 2

    grid.set_watershed_boundary_condition_outlet_id(outlet_id, z,
                                                    nodata_value=-9999.)

    fr = FlowRouter(grid)
    fr.run_one_step()

    ws_nodes = get_watershed_nodes(grid, outlet_id)

    # Given the watershed boundary conditions, the number of watershed nodes
    # should be equal to the number of core nodes plus 1 for the outlet node.
    np.testing.assert_equal(len(ws_nodes), grid.number_of_core_nodes + 1)


def test_get_watershed_masks_with_area_threshold():
    rmg = RasterModelGrid((7, 7), 200)

    z = np.array([
            -9999., -9999., -9999., -9999., -9999., -9999., -9999.,
            -9999.,    26.,     0.,    26.,    30.,    34., -9999.,
            -9999.,    28.,     1.,    28.,     5.,    32., -9999.,
            -9999.,    30.,     3.,    30.,    10.,    34., -9999.,
            -9999.,    32.,    11.,    32.,    15.,    38., -9999.,
            -9999.,    34.,    32.,    34.,    36.,    40., -9999.,
            -9999., -9999., -9999., -9999., -9999., -9999., -9999.])

    rmg.at_node['topographic__elevation'] = z

    rmg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    # Route flow.
    fr = FlowRouter(rmg)
    fr.run_one_step()

    # Get the masks of watersheds greater than or equal to 80,000
    # square-meters.
    critical_area = 80000
    mask = get_watershed_masks_with_area_threshold(rmg, critical_area)

    # Assert that mask null nodes have a drainage area below critical area.
    null_nodes = np.where(mask == -1)[0]
    A = rmg.at_node['drainage_area'][null_nodes]
    below_critical_area_nodes = A < critical_area
    trues = np.ones(len(A), dtype=bool)

    np.testing.assert_equal(below_critical_area_nodes, trues)


def test_get_watershed_outlet():
    grid = RasterModelGrid((7, 7), 1)

    z = np.array([
        -9999., -9999., -9999., -9999., -9999., -9999., -9999.,
        -9999.,    26.,     0.,    30.,    32.,    34., -9999.,
        -9999.,    28.,     1.,    25.,    28.,    32., -9999.,
        -9999.,    30.,     3.,     3.,    11.,    34., -9999.,
        -9999.,    32.,    11.,    25.,    18.,    38., -9999.,
        -9999.,    34.,    32.,    34.,    36.,    40., -9999.,
        -9999., -9999., -9999., -9999., -9999., -9999., -9999.])

    grid.at_node['topographic__elevation'] = z

    imposed_outlet = 2

    grid.set_watershed_boundary_condition_outlet_id(imposed_outlet, z,
                                                    nodata_value=-9999.)

    fr = FlowRouter(grid)
    fr.run_one_step()

    test_node = 32

    determined_outlet = get_watershed_outlet(grid, test_node)
    np.testing.assert_equal(determined_outlet, imposed_outlet)

    # Create a pit.
    pit_node = 38
    grid.at_node['topographic__elevation'][pit_node] -= 32
    fr.run_one_step()

    pit_outlet = get_watershed_outlet(grid, test_node)
    np.testing.assert_equal(pit_outlet, pit_node)



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 14 10:36:03 2015

@author: gtucker
"""
import numpy as np

from landlab import HexModelGrid
from numpy.testing import assert_array_equal


def test_hex_grid_link_order():
    """Test the order of links in a small hex grid."""
    hg = HexModelGrid(3, 2)
    assert_array_equal(hg.node_at_link_tail, [0, 0, 0, 1, 1, 2, 3, 2, 3, 3,
                                              4, 5])
    assert_array_equal(hg.node_at_link_head, [1, 2, 3, 3, 4, 3, 4, 5, 5, 6,
                                              6, 6])

    hg = HexModelGrid(2, 3, orientation='vertical')
    assert_array_equal(hg.node_at_link_tail, [1, 0, 0, 1, 3, 1, 2, 4, 3, 3,
                                              4, 6])
    assert_array_equal(hg.node_at_link_head, [0, 2, 3, 3, 2, 4, 5, 3, 5, 6,
                                              6, 5])


def test_nodes_at_link():
    """Test nodes_at_link shares data with tail and head."""
    grid = HexModelGrid(3, 2)

    assert_array_equal(grid.nodes_at_link[:, 0], grid.node_at_link_tail)
    assert_array_equal(grid.nodes_at_link[:, 1], grid.node_at_link_head)

    assert np.may_share_memory(grid.nodes_at_link, grid.node_at_link_tail)
    assert np.may_share_memory(grid.nodes_at_link, grid.node_at_link_head)


if __name__=='__main__':
    test_hex_grid_link_order()



import numpy as np
cimport numpy as np
cimport cython

DTYPE = np.int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def calc_midpoint_of_link(np.ndarray[DTYPE_t, ndim=2] nodes_at_link,
                          np.ndarray[np.float_t, ndim=1] x_of_node,
                          np.ndarray[np.float_t, ndim=1] y_of_node,
                          np.ndarray[np.float_t, ndim=2] xy_of_link):
    cdef int link
    cdef int n_links = nodes_at_link.shape[0]

    for link in range(n_links):
        link_tail = nodes_at_link[link][0]
        link_head = nodes_at_link[link][1]

        xy_of_link[link][0] = (x_of_node[link_tail] +
                               x_of_node[link_head]) * .5
        xy_of_link[link][1] = (y_of_node[link_tail] +
                               y_of_node[link_head]) * .5



from six import StringIO

from landlab.core import load_params


YAML_PARAMS_STR = """
x: 1e7
y: 1
z: [1, 2]
a: frog
"""
MPD_PARAMS_STR = """
x: A float
1e7
y: An int
1
a: A string
frog
"""
YAML_PARAMS = {
    'x': 1e7,
    'y': 1,
    'z': [1, 2],
    'a': 'frog',
}
MPD_PARAMS = {
    'x': 1e7,
    'y': 1,
    'a': 'frog',
}


def test_from_yaml_string():
    """Load parameters from YAML-formatted string."""
    params = load_params(YAML_PARAMS_STR)
    assert params == YAML_PARAMS
    assert isinstance(params['x'], float)
    assert isinstance(params['y'], int)


def test_from_yaml_file_like():
    """Load parameters from YAML-formatted string."""
    params = load_params(StringIO(YAML_PARAMS_STR))
    assert params == YAML_PARAMS
    assert isinstance(params['x'], float)
    assert isinstance(params['y'], int)


def test_from_yaml_path(tmpdir):
    """Load parameters from YAML-formatted string."""
    with tmpdir.as_cwd():
        with open('params.yaml', 'w') as fp:
            fp.write(YAML_PARAMS_STR)
        params = load_params('./params.yaml')
    assert params == YAML_PARAMS
    assert isinstance(params['x'], float)
    assert isinstance(params['y'], int)


def test_from_mpd_string():
    """Load parameters from YAML-formatted string."""
    params = load_params(MPD_PARAMS_STR)
    assert params == MPD_PARAMS
    assert isinstance(params['x'], float)
    assert isinstance(params['y'], int)


def test_from_yaml_file_like():
    """Load parameters from YAML-formatted string."""
    params = load_params(StringIO(MPD_PARAMS_STR))
    assert params == MPD_PARAMS
    assert isinstance(params['x'], float)
    assert isinstance(params['y'], int)


def test_from_yaml_path(tmpdir):
    """Load parameters from YAML-formatted string."""
    with tmpdir.as_cwd():
        with open('params.txt', 'w') as fp:
            fp.write(MPD_PARAMS_STR)
        params = load_params('./params.txt')
    assert params == MPD_PARAMS
    assert isinstance(params['x'], float)
    assert isinstance(params['y'], int)



#!/usr/bin/env python
"""Deform the lithosphere with 1D or 2D flexure.

Landlab component that implements a 1D lithospheric flexure.

Examples
--------

Create a grid on which we will run the flexure calculations.

>>> from landlab import RasterModelGrid
>>> from landlab.components.flexure import Flexure1D
>>> grid = RasterModelGrid((3, 4), spacing=(1.e4, 1.e4))

Because `Flexure1D` is a one-dimensional component, it operates
*independently* on each row of grid nodes. By default, it will
calculate deflections on every row but you can also specify
which rows to operate on by passing an indexing array to the
`rows` keyword.

In this example, we'll just calculate deflections along the
middle row or nodes (that is, row 1)

>>> flex = Flexure1D(grid, rows=1)

In creating the component, a field (initialized with zeros) was
added to the grid that represents the current loading distribution.
If the grid already had this field, the component would use the
existing field. This can be accessed either through the *grid*
attribute in the same way as other landlab fields,

>>> flex.grid.at_node['lithosphere__increment_of_overlying_pressure']
array([ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.])

or through the `load_at_node` attribute of `Flexure1D`,

>>> flex.load_at_node
array([[ 0.,  0.,  0.,  0.],
       [ 0.,  0.,  0.,  0.],
       [ 0.,  0.,  0.,  0.]])

Notice that `load_at_node` returns a reshaped view of the array
whereas the field returns a flattened array. Change values in this
array to add loads to the grid,

>>> flex.load_at_node[1, 2:4] = flex.gamma_mantle
>>> flex.run_one_step()

The output deflections can be retrieved either using landlab fields
as,

>>> flex.grid.at_node['lithosphere_surface__increment_of_elevation']
array([ 0.,  0.,  0.,  0.,  0.,  0.,  1.,  1.,  0.,  0.,  0.,  0.])

or through the `dz_at_node` attribute,

>>> flex.dz_at_node
array([[ 0.,  0.,  0.,  0.],
       [ 0.,  0.,  1.,  1.],
       [ 0.,  0.,  0.,  0.]])
"""

import numpy as np

from landlab import Component
from .funcs import get_flexure_parameter
from ...utils.decorators import use_file_name_or_kwds
from .ext import subside_load_1d


class Flexure1D(Component):

    """Deform the lithosphere with 1D flexure.

    Landlab component that implements a 1D lithospheric flexure model.

    Parameters
    ----------
    grid : RasterModelGrid
        A grid.
    eet : float, optional
        Effective elastic thickness (m).
    youngs : float, optional
        Young's modulus.
    method : {'airy', 'flexure'}, optional
        Method to use to calculate deflections.
    rho_mantle : float, optional
        Density of the mantle (kg / m^3).
    rho_water : float, optional
        Density of the sea water (kg / m^3).
    gravity : float, optional
        Acceleration due to gravity (m / s^2).
    rows : int, optional
        Node rows that this component will operate on (default is to
        operate on *all* rows).

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.flexure import Flexure
    >>> grid = RasterModelGrid((5, 4), spacing=(1.e4, 1.e4))

    >>> flex = Flexure(grid)
    >>> flex.name
    'Flexure'
    >>> flex.input_var_names
    ('lithosphere__overlying_pressure_increment',)
    >>> flex.output_var_names
    ('lithosphere_surface__elevation_increment',)
    >>> sorted(flex.units) # doctest: +NORMALIZE_WHITESPACE
    [('lithosphere__overlying_pressure_increment', 'Pa'),
     ('lithosphere_surface__elevation_increment', 'm')]

    >>> flex.grid.number_of_node_rows
    5
    >>> flex.grid.number_of_node_columns
    4
    >>> flex.grid is grid
    True

    >>> np.all(grid.at_node['lithosphere_surface__elevation_increment'] == 0.)
    True

    >>> np.all(grid.at_node['lithosphere__overlying_pressure_increment'] == 0.)
    True
    >>> flex.update()
    >>> np.all(grid.at_node['lithosphere_surface__elevation_increment'] == 0.)
    True

    >>> load = grid.at_node['lithosphere__overlying_pressure_increment']
    >>> load[4] = 1e9
    >>> dz = grid.at_node['lithosphere_surface__elevation_increment']
    >>> np.all(dz == 0.)
    True

    >>> flex.update()
    >>> np.all(grid.at_node['lithosphere_surface__elevation_increment'] == 0.)
    False
    """

    _name = "1D Flexure Equation"

    _input_var_names = ("lithosphere__increment_of_overlying_pressure",)

    _output_var_names = ("lithosphere_surface__increment_of_elevation",)

    _var_units = {
        "lithosphere__increment_of_overlying_pressure": "Pa",
        "lithosphere_surface__increment_of_elevation": "m",
    }

    _var_mapping = {
        "lithosphere__increment_of_overlying_pressure": "node",
        "lithosphere_surface__increment_of_elevation": "node",
    }

    _var_doc = {
        "lithosphere__increment_of_overlying_pressure": "Applied pressure to the lithosphere over a time step",
        "lithosphere_surface__increment_of_elevation": "The change in elevation of the top of the lithosphere (the land "
        "surface) in one timestep",
    }

    POISSON = .25

    @use_file_name_or_kwds
    def __init__(
        self,
        grid,
        eet=65e3,
        youngs=7e10,
        method="airy",
        rho_mantle=3300.,
        rho_water=1030.,
        gravity=9.80665,
        rows=None,
        **kwds
    ):
        """Initialize the flexure component.

        Parameters
        ----------
        grid : RasterModelGrid
            A grid.
        eet : float, optional
            Effective elastic thickness (m).
        youngs : float, optional (Pa)
            Young's modulus.
        method : {'airy', 'flexure'}, optional
            Method to use to calculate deflections.
        rho_mantle : float, optional
            Density of the mantle (kg / m^3).
        rho_water : float, optional
            Density of the sea water (kg / m^3).
        gravity : float, optional
            Acceleration due to gravity (m / s^2).
        rows : int, optional
            Node rows that this component will operate on (default is to
            operate on *all* rows).
        """
        if method not in ("airy", "flexure"):
            raise ValueError("{method}: method not understood".format(method=method))

        self._grid = grid

        self._method = method
        self.youngs = youngs
        self.rho_mantle = rho_mantle
        self.rho_water = rho_water
        self.gravity = gravity
        self.eet = eet

        super(Flexure1D, self).__init__(grid, **kwds)

        for name in self._input_var_names + self._output_var_names:
            if name not in self.grid.at_node:
                self.grid.add_zeros(name, units=self._var_units[name], at="node")

        self._rows = (rows,) or Ellipsis

        self._x_at_node = self.grid.x_of_node.reshape(self.grid.shape).copy()

    @property
    def eet(self):
        """Effective elastic thickness (m)."""
        return self._eet

    @eet.setter
    def eet(self, new_val):
        if new_val <= 0:
            raise ValueError("Effective elastic thickness must be positive.")
        try:
            del self._rigidity, self._alpha
        except AttributeError:
            pass
        self._eet = float(new_val)

    @property
    def youngs(self):
        """Young's modulus of lithosphere (Pa)."""
        return self._youngs

    @youngs.setter
    def youngs(self, new_val):
        if new_val <= 0:
            raise ValueError("Young's modulus must be positive.")
        try:
            del self._rigidity, self._alpha
        except AttributeError:
            pass
        self._youngs = float(new_val)

    @property
    def rho_water(self):
        """Density of water (kg/m^3)."""
        return self._rho_water

    @rho_water.setter
    def rho_water(self, new_val):
        if new_val <= 0:
            raise ValueError("Water density must be positive.")
        try:
            del self._gamma_mantle, self._alpha
        except AttributeError:
            pass
        self._rho_water = float(new_val)

    @property
    def rho_mantle(self):
        """Density of mantle (kg/m^3)."""
        return self._rho_mantle

    @rho_mantle.setter
    def rho_mantle(self, new_val):
        if new_val <= 0:
            raise ValueError("Mantle density must be positive.")
        try:
            del self._gamma_mantle, self._alpha
        except AttributeError:
            pass
        self._rho_mantle = float(new_val)

    @property
    def gravity(self):
        """Acceleration due to gravity (m/s^2)."""
        return self._gravity

    @gravity.setter
    def gravity(self, new_val):
        if new_val <= 0:
            raise ValueError("Acceleration due to gravity must be positive.")
        try:
            del self._gamma_mantle, self._alpha
        except AttributeError:
            pass
        self._gravity = float(new_val)

    @property
    def alpha(self):
        """Flexure parameter (m)."""
        try:
            self._alpha
        except AttributeError:
            self._alpha = np.power(4 * self.rigidity / self.gamma_mantle, .25)
        finally:
            return self._alpha

    @property
    def rigidity(self):
        """Flexural rigidity (N m)."""
        try:
            self._rigidity
        except AttributeError:
            self._rigidity = (
                self._eet ** 3. * self._youngs / (12. * (1. - self.POISSON ** 2.))
            )
        finally:
            return self._rigidity

    @property
    def gamma_mantle(self):
        """Specific density of mantle (N/m^3)."""
        try:
            self._gamma_mantle
        except AttributeError:
            self._gamma_mantle = (self._rho_mantle - self._rho_water) * self._gravity
        finally:
            return self._gamma_mantle

    @property
    def method(self):
        """Name of method used to calculate deflections."""
        return self._method

    @property
    def x_at_node(self):
        return self._x_at_node

    @property
    def load_at_node(self):
        return self.grid.at_node[
            "lithosphere__increment_of_overlying_pressure"
        ].reshape(self.grid.shape)

    @property
    def dz_at_node(self):
        return self.grid.at_node["lithosphere_surface__increment_of_elevation"].reshape(
            self.grid.shape
        )

    def update(self):
        """Update fields with current loading conditions."""
        load = self.load_at_node[self._rows]
        deflection = self.dz_at_node[self._rows]

        if self._method == "airy":
            deflection[:] = load / self.gamma_mantle
        else:
            Flexure1D.calc_flexure(
                self.x_at_node[0], load, self.alpha, self.rigidity, out=deflection
            )

        if not np.may_share_memory(deflection, self.dz_at_node):
            self.dz_at_node[self._rows] = deflection

    def run_one_step(self, dt=None):
        self.update()

    def subside_loads(self, loads, out=None):
        """Subside surface due to multiple loads.

        Parameters
        ----------
        loads : ndarray of float
            Loads applied to each grid node (Pa).
        out : ndarray of float, optional
            Buffer to place resulting deflection values (m).

        Returns
        -------
        ndarray of float
            Deflections caused by the loading.
        """
        if self._method == "airy":
            deflection[:] = load / self.gamma_mantle
        else:
            Flexure1D.calc_flexure(
                self.x_at_node[0], loads, self.alpha, self.rigidity, out=out
            )

        return out

    @staticmethod
    def calc_flexure(x, loads, alpha, rigidity, out=None):
        """Subside surface due to multiple loads.

        Parameters
        ----------
        x : ndarray of float
            Position of grid nodes (m).
        loads : ndarray of float
            Loads applied to each grid node (Pa).
        alpha : float
            Flexure parameter of the lithosphere (m).
        rigidity : float
            Flexural rigidity of the lithosphere (N m).
        out : ndarray of float, optional
            Buffer to place resulting deflection values (m).

        Returns
        -------
        ndarray of float
            Deflections caused by the loading.
        """
        if out is None:
            out = np.zeros_like(loads, dtype=np.float)

        loads = loads.reshape((-1, loads.shape[-1]))
        out = out[..., :].reshape((-1, out.shape[-1]))
        subside_load_1d(x, loads, alpha, rigidity, out[..., :])

        if out.base is not None:
            return out.base
        else:
            return out



"""Sort the elements of a graph.

This module provides functions that sort the elements of a graph structure.
"""

import numpy as np

from ...core.utils import as_id_array, argsort_points_by_x_then_y
from ...utils.jaggedarray import flatten_jagged_array, unravel
from .ext.spoke_sort import sort_spokes_at_wheel


def remap(src, mapping, out=None, inplace=False): 
    """Remap elements in an id array.

    Parameters
    ----------
    src : ndarray of int
        Initial array of ids.
    mapping : ndarray of int
        Mapping of ids.
    out : ndarray of int, optional
        Buffer into which to place output.
    inplace : bool, optional
        Mapping of values will include inplace.

    Returns
    -------
    ndarray of int
        Array of remapped values.

    Examples
    --------
    >>> from landlab.graph.sort.sort import remap
    >>> import numpy as np

    >>> src = np.array([1, 2, 3, 4])
    >>> mapping = np.array([-1, 10, 20, 30, 40])
    >>> remap(src, mapping)
    array([10, 20, 30, 40])
    """
    from .ext.remap_element import remap_graph_element

    if inplace:
        out = src
    else:
        if out is None:
            out = src.copy()
        else:
            out[:] = src[:]
    
    remap_graph_element(out.reshape((-1, )), mapping)

    return out


def reverse_one_to_one(ids, minlength=None):
    """Reverse a one-to-one mapping.

    Parameters
    ----------
    ids : ndarray of int, shape `(N, )`
        Array of identifier mapping.
    minlength : int, optional
        A minimum number of identifiers for the output array.
    Returns
    -------
    ndarray of int, shape `(n, )`
        Array of the reverse mapping.

    Examples
    --------
    >>> from landlab.graph.sort.sort import reverse_one_to_one
    >>> ids = np.array([-1, -1, 6, 3, -1, 2, 4, 1, -1, 5, 0], dtype=int)
    >>> reverse_one_to_one(ids)
    array([10,  7,  5,  3,  6,  9,  2])
    """
    from .ext.remap_element import reverse_one_to_one

    if minlength is None:
        minlength = ids.max() + 1
    out = np.full((minlength, ), -1, dtype=int)

    reverse_one_to_one(ids, out)

    return out


def reverse_one_to_many(ids, min_counts=0):
    """Reverse a one-to-many mapping.

    Parameters
    ----------
    ids : ndarray of int, shape `(M, N)`
        Array of identifier mapping.

    Returns
    -------
    ndarray of int, shape `(m, n)`
        Array of the reverse mapping.

    Examples
    --------
    >>> from landlab.graph.sort.sort import reverse_one_to_many
    >>> ids = np.array([[1,2,3],[-1,-1,-1],[2,3,-1]], dtype=int)
    >>> reverse_one_to_many(ids)
    array([[-1, -1],
           [ 0, -1],
           [ 0,  2],
           [ 0,  2]])
    """
    from .ext.remap_element import reverse_one_to_many

    counts = np.bincount(ids.reshape((-1, )) + 1)
    max_counts = np.max((np.max(counts[1:]), min_counts))

    out = np.full((ids.max() + 1, max_counts), -1, dtype=int)

    reverse_one_to_many(ids, out)

    return out


def reorder_links_at_patch(graph):
    from ..quantity.of_patch import get_area_of_patch
    from ..quantity.of_link import get_midpoint_of_link
    from ..matrix.ext.matrix import roll_id_matrix_rows
    from ..object.ext.at_patch import get_rightmost_edge_at_patch
    from .ext.remap_element import reverse_element_order
    from ..object.at_patch import get_nodes_at_patch

    if graph.number_of_patches == 0:
        return

    xy_of_link = get_midpoint_of_link(graph)

    shift = np.empty(graph.number_of_patches, dtype=int)

    get_rightmost_edge_at_patch(graph.links_at_patch, xy_of_link, shift)
    roll_id_matrix_rows(graph.links_at_patch, - shift)

    before = graph.links_at_patch.copy()
    area_before = get_area_of_patch(graph)

    negative_areas = as_id_array(np.where(get_area_of_patch(graph) < 0.)[0])
    reverse_element_order(graph.links_at_patch, negative_areas)
    # reverse_element_order(graph._links_at_patch, negative_areas)

    # graph._nodes_at_patch = get_nodes_at_patch(graph)
    if 'nodes_at_patch' in graph._ds:
        graph._ds = graph._ds.drop('nodes_at_patch')

    if np.any(get_area_of_patch(graph) < 0.):
        raise ValueError((graph.links_at_patch,
                          before,
                          get_area_of_patch(graph),
                          area_before))


def reorient_link_dirs(graph):
    from ..quantity.of_link import get_angle_of_link

    if graph.number_of_links == 0:
        return

    angles = get_angle_of_link(graph)
    links_to_swap = (angles < 7. * np.pi / 4.) & (angles > np.pi * .75)
    graph.nodes_at_link[links_to_swap, :] = (
        graph.nodes_at_link[links_to_swap, ::-1])


def reindex_by_xy(graph):
    sorted_nodes = reindex_nodes_by_xy(graph)
    # if hasattr(graph, '_nodes_at_link'):
    if 'nodes_at_link' in graph.ds:
        sorted_links = reindex_links_by_xy(graph)
    else:
        sorted_links = None

    # if hasattr(graph, '_links_at_patch'):
    if 'links_at_patch' in graph.ds:
        sorted_patches = reindex_patches_by_xy(graph)
    else:
        sorted_patches = None

    return sorted_nodes, sorted_links, sorted_patches


def reindex_patches_by_xy(graph):
    from ..quantity.of_patch import get_centroid_of_patch

    if graph.number_of_patches == 1:
        return np.array([0], dtype=int)

    xy_at_patch = get_centroid_of_patch(graph)
    xy_at_patch[:, 1] = np.round(xy_at_patch[:, 1], decimals=5)

    sorted_patches = argsort_points_by_x_then_y(
        (xy_at_patch[:, 0], xy_at_patch[:, 1]))

    graph.links_at_patch[:] = graph.links_at_patch[sorted_patches, :]
    # graph._links_at_patch[:] = graph._links_at_patch[sorted_patches, :]

    if 'nodes_at_patch' in graph._ds:
        graph._ds = graph.ds.drop('nodes_at_patch')
    # del graph.__dict__['_nodes_at_patch']

    # if hasattr(graph, '_node_at_cell'):
    #     graph._node_at_cell[:] = graph._node_at_cell[sorted_patches]

    return sorted_patches


def reindex_links_by_xy(graph):
    from ..quantity.of_link import get_midpoint_of_link
    from .ext.remap_element import remap_graph_element_ignore

    xy_of_link = get_midpoint_of_link(graph)

    sorted_links = argsort_points_by_x_then_y(xy_of_link)

    # graph._nodes_at_link[:] = graph._nodes_at_link[sorted_links, :]
    graph.nodes_at_link[:] = graph.nodes_at_link[sorted_links, :]

    # if hasattr(graph, '_links_at_patch'):
    if 'links_at_patch' in graph.ds:
        remap_graph_element_ignore(graph.links_at_patch.reshape((-1, )),
                                   as_id_array(np.argsort(sorted_links)), -1)

    return sorted_links


def reindex_nodes_by_xy(graph):
    from .ext.remap_element import remap_graph_element

    graph.y_of_node[:] = np.round(graph.y_of_node, decimals=6)

    sorted_nodes = argsort_points_by_x_then_y((graph.x_of_node,
                                               graph.y_of_node))

    graph.y_of_node[:] = graph.y_of_node[sorted_nodes]
    graph.x_of_node[:] = graph.x_of_node[sorted_nodes]

    if 'nodes_at_link' in graph.ds:
        remap_graph_element(graph.nodes_at_link.reshape((-1, )),
                            as_id_array(np.argsort(sorted_nodes)))

    if 'nodes_at_patch' in graph.ds:
        remap_graph_element(graph.nodes_at_patch.reshape((-1, )),
                            as_id_array(np.argsort(sorted_nodes)))

    return sorted_nodes


def sort_graph(nodes, links=None, patches=None):
    """Sort elements of a graph by x, then y.

    Parameters
    ----------
    nodes : tuple of ndarray
        Coordinate of nodes as (y, x).
    links : ndarray of int, shape `(n_links, 2)`, optional
        Indices into *nodes* array of link tail then head.
    patches : array_like of array_like, optional
        Indices into *links* array of the links that define each patch.

    Returns
    -------
    (nodes, links, patches)
        Sorted nodes, links, and patches.

    Examples
    --------
    o---o---o
    |   | / |
    o---o---o

    >>> from landlab.graph.sort import sort_graph
    >>> import numpy as np
    >>> x = np.array([1., 2., 2., 0., 1., 0.])
    >>> y = np.array([0., 0., 1., 0., 1., 1.])

    Sort a graph with just points - no links or patches.

    >>> _ = sort_graph((y, x))
    >>> y
    array([ 0.,  0.,  0.,  1.,  1.,  1.])
    >>> x
    array([ 0.,  1.,  2.,  0.,  1.,  2.])

    Sort the points and links of a graph.

    >>> x = np.array([1., 2., 2., 0., 1., 0.])
    >>> y = np.array([0., 0., 1., 0., 1., 1.])
    >>> links = np.array([[3, 0], [0, 4], [4, 5], [5, 3], [0, 1], [1, 2],
    ...                   [2, 0], [2, 4]])
    >>> _ = sort_graph((y, x), links)
    >>> links # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1], [1, 2], [3, 0], [1, 4], [5, 1], [2, 5], [4, 3], [5, 4]])

    Sort the points, links, and patches of a graph.

    >>> x = np.array([1., 2., 2., 0., 1., 0.])
    >>> y = np.array([0., 0., 1., 0., 1., 1.])
    >>> links = np.array([[3, 0], [0, 4], [4, 5], [5, 3], [0, 1], [1, 2],
    ...                   [2, 0], [2, 4]])
    >>> patches = (np.array([1, 6, 7, 4, 5, 6, 0, 1, 2, 3]),
    ...            np.array([0, 3, 6, 10]))
    >>> _ = sort_graph((y, x), links, patches)
    >>> patches[0]
    array([1, 5, 4, 0, 3, 6, 2, 3, 4, 7])
    >>> patches[1]
    array([ 0,  3,  7, 10])
    """
    from .ext.remap_element import remap_graph_element

    if patches is not None and links is None:
        raise ValueError('graph that has patches must also have links')

    if links is not None:
        links = as_id_array(links)

    if patches is not None:
        if len(patches) == 2 and isinstance(patches[0], np.ndarray):
            links_at_patch, offset_to_patch = patches
        else:
            links_at_patch, offset_to_patch = flatten_jagged_array(patches,
                                                                   dtype=int)
        links_at_patch, offset_to_patch = (as_id_array(links_at_patch),
                                           as_id_array(offset_to_patch))
    else:
        links_at_patch, offset_to_patch = (None, None)

    sorted_nodes = sort_nodes(nodes)

    if links is not None:
        remap_graph_element(links.reshape((-1, )),
                            as_id_array(np.argsort(sorted_nodes,
                                                   kind='mergesort')))
        midpoint_of_link = np.empty((len(links), 2), dtype=float)
        sorted_links = sort_links(links, nodes,
                                  midpoint_of_link=midpoint_of_link) 

    if patches is not None:
        remap_graph_element(links_at_patch,
                            as_id_array(np.argsort(sorted_links,
                                                   kind='mergesort')))
        sort_patches(links_at_patch, offset_to_patch, midpoint_of_link)

    if links_at_patch is None:
        return nodes, links, None
    else:
        return nodes, links, (links_at_patch, offset_to_patch)


def sort_nodes(nodes):
    """Sort nodes based on their position.

    Parameters
    ----------
    nodes : tuple of ndarray of float
        Coordinates of nodes as (*y*, *x*).

    Returns
    -------
    ndarray of int
        Array of indices that sort the nodes.

    Examples
    --------
    >>> from landlab.graph.sort import sort_nodes
    >>> import numpy as np
    >>> x = np.array([0. , 1., 2.])
    >>> y = np.array([ .5, 0., 1.])
    >>> sort_nodes((y, x))
    array([1, 0, 2])
    >>> x
    array([ 1.,  0.,  2.])
    >>> y
    array([ 0. ,  0.5,  1. ])
    """
    sorted_nodes = argsort_points_by_x_then_y((nodes[1], nodes[0]))
    nodes[0][:] = nodes[0][sorted_nodes]
    nodes[1][:] = nodes[1][sorted_nodes]

    return sorted_nodes


def sort_links(nodes_at_link, nodes, midpoint_of_link=None):
    """Sort links by their midpoint.

    Parameters
    ----------
    nodes_at_link : ndarray of int, shape `(n_links, 2)`
        Node for each link tail and head.
    nodes : tuple of ndarray of float
        Node coordinates.
    midpoint_of_link : ndarray of float, shape `(n_links, 2)`, optional
        Buffer to store the link midpoints that were used for sorting.

    Returns
    -------
    ndarray of int
        Array of indices that sort the links.

    Examples
    --------
    >>> from landlab.graph.sort import sort_nodes
    >>> import numpy as np
    >>> nodes = np.array([[0, 0, 0, 1, 1, 1],
    ...                   [0, 1, 2, 0, 1, 2]])
    >>> links = np.array([[0, 1], [0, 3], [1, 2], [1, 4], [2, 5], [3, 4],
    ...                   [4, 5]])
    >>> sort_links(links, nodes)
    array([0, 2, 1, 3, 4, 5, 6])
    >>> links
    array([[0, 1],
           [1, 2],
           [0, 3],
           [1, 4],
           [2, 5],
           [3, 4],
           [4, 5]])
    """
    from ..quantity.ext.of_link import calc_midpoint_of_link

    y_of_node, x_of_node = np.asfarray(nodes[0]), np.asfarray(nodes[1])

    if midpoint_of_link is None:
        midpoint_of_link = np.empty((len(nodes_at_link), 2), dtype=float)

    calc_midpoint_of_link(nodes_at_link, x_of_node, y_of_node,
                          midpoint_of_link)

    sorted_links = argsort_points_by_x_then_y(midpoint_of_link)
    nodes_at_link[:] = nodes_at_link[sorted_links]
    midpoint_of_link[:] = midpoint_of_link[sorted_links]

    return sorted_links


def sort_patches(links_at_patch, offset_to_patch, xy_of_link):
    """Sort patches by their centroid.

    Parameters
    ----------
    links_at_patch : ndarray of int
        Links that define patches.
    offset_to_patch : ndarray of int
        Offsets into *links_at_patch* for each patch.
    xy_of_link : ndarray of float, shape `(n_links, 2)`
        Midpoint coordinates for each link.

    Examples
    --------
    >>> from landlab.graph.sort import sort_patches
    >>> import numpy as np
    >>> links_at_patch = np.array([0, 1, 2, 3, 2, 4])
    >>> offset_to_patch = np.array([0, 3, 6])
    >>> xy_of_link = np.array([[0.0, 0.5], [0.5, 1.0], [.5, .5],
    ...                        [0.5, 0.0], [1.0, 0.5]])
    >>> sort_patches(links_at_patch, offset_to_patch, xy_of_link)
    array([1, 0])
    >>> links_at_patch
    array([3, 2, 4, 0, 1, 2])
    >>> offset_to_patch
    array([0, 3, 6])
    """
    from .ext.remap_element import calc_center_of_patch, reorder_patches

    n_patches = len(offset_to_patch) - 1
    xy_at_patch = np.empty((n_patches, 2), dtype=float)

    calc_center_of_patch(links_at_patch, offset_to_patch,
                         xy_of_link, xy_at_patch)

    sorted_patches = argsort_points_by_x_then_y(xy_at_patch)
    reorder_patches(links_at_patch, offset_to_patch, sorted_patches)

    return sorted_patches


def sort_spokes_at_hub(graph, spoke=None, at='node', inplace=False):
    """Order spokes of a graph clockwise around spokes.

    Parameters
    ----------
    graph : Graph-like
        A landlab graph.
    spoke : str
        Name of graph elements that make the spokes.
    at : {'node', 'corner', 'link', 'face', 'patch', 'cell'}
        Namve of graph elements that make the hubs.

    Returns
    -------
    ndarray of int, shape `(n_spokes, n_hubs)`
        Spokes ordered around each hub.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.graph import UniformRectilinearGraph
    >>> from landlab.graph.sort.sort import sort_spokes_at_hub

    >>> graph = UniformRectilinearGraph((3, 3))
    >>> sort_spokes_at_hub(graph, 'link', at='node')
    array([[ 0,  2, -1, -1],
           [ 1,  3,  0, -1],
           [ 4,  1, -1, -1],
           [ 5,  7,  2, -1],
           [ 6,  8,  5,  3],
           [ 9,  6,  4, -1],
           [10,  7, -1, -1],
           [11, 10,  8, -1],
           [11,  9, -1, -1]])
    """
    sorted_spokes = argsort_spokes_at_hub(graph, spoke=spoke, at=at)
    if spoke == 'patch':
        plural = 'patches'
    else:
        plural = spoke + 's'
    spokes_at_hub = getattr(graph,
                            '{plural}_at_{hub}'.format(plural=plural, hub=at))

    if inplace:
        out = spokes_at_hub
    else:
        out = np.empty_like(spokes_at_hub)

    return np.take(spokes_at_hub, sorted_spokes, out=out)


def argsort_spokes_at_hub(graph, spoke=None, at='node'):
    """Order spokes clockwise around spokes.

    Parameters
    ----------
    graph : Graph-like
        A landlab graph.
    spoke : str
        Name of graph elements that make the spokes.
    at : str
        Namve of graph elements that make the hubs.

    Returns
    -------
    ndarray of int, shape `(n_spokes, n_hubs)`
        Spokes ordered around each hub.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.graph import UniformRectilinearGraph
    >>> from landlab.graph.sort.sort import argsort_spokes_at_hub

    >>> graph = UniformRectilinearGraph((3, 3))
    >>> argsort_spokes_at_hub(graph, 'link', at='node')
    array([[ 0,  1,  2,  3],
           [ 4,  5,  6,  7],
           [ 9, 10,  8, 11],
           [12, 13, 15, 14],
           [16, 17, 18, 19],
           [21, 22, 23, 20],
           [24, 27, 25, 26],
           [28, 30, 31, 29],
           [34, 35, 32, 33]])
    """
    angles = calc_angle_of_spoke(graph, spoke=spoke, at=at, badval=np.inf)
    angles[angles < 0] += np.pi

    n_hubs, n_spokes = angles.shape
    ordered_angles = np.argsort(angles)
    ordered_angles += np.arange(n_hubs).reshape((-1, 1)) * n_spokes

    return as_id_array(ordered_angles)


def calc_angle_of_spoke(graph, spoke=None, at='node', badval=None):
    """Calculate angles spokes make with a hub.

    Parameters
    ----------
    graph : Graph-like
        A landlab graph.
    spoke : str
        Name of graph elements that make the spokes.
    at : str
        Namve of graph elements that make the hubs.
    badval : float or iterable of float, optional
        Value to insert for missing spokes. If an iterable, use items as
        bad values to use for different spokes.

    Returns
    -------
    ndarray of float, shape `(n_spokes, n_hubs)`
        Angle that spoke elements make with each hub element.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.graph import UniformRectilinearGraph
    >>> from landlab.graph.sort.sort import calc_angle_of_spoke

    >>> graph = UniformRectilinearGraph((3, 3))
    >>> np.rad2deg(calc_angle_of_spoke(graph, 'link', at='node', badval=np.nan))
    array([[   0.,   90.,   nan,   nan],
           [   0.,   90.,  180.,   nan],
           [  nan,   90.,  180.,   nan],
           [   0.,   90.,   nan,  270.],
           [   0.,   90.,  180.,  270.],
           [  nan,   90.,  180.,  270.],
           [   0.,   nan,   nan,  270.],
           [   0.,   nan,  180.,  270.],
           [  nan,   nan,  180.,  270.]])
    """
    xy_of_hub = getattr(graph, 'xy_of_{hub}'.format(hub=at))
    xy_of_spoke = getattr(graph, 'xy_of_{spoke}'.format(spoke=spoke))
    if spoke == 'patch':
        plural = 'patches'
    else:
        plural = spoke + 's'
    spokes_at_hub = getattr(graph,
                            '{plural}_at_{hub}'.format(plural=plural, hub=at))

    xy_of_spoke = xy_of_spoke[spokes_at_hub.flat]
    x_of_spoke = xy_of_spoke[:, 0].reshape(spokes_at_hub.shape)
    y_of_spoke = xy_of_spoke[:, 1].reshape(spokes_at_hub.shape)

    x_of_hub, y_of_hub = xy_of_hub[:, 0], xy_of_hub[:, 1]

    dx = (x_of_spoke.T - x_of_hub).T
    dy = (y_of_spoke.T - y_of_hub).T

    angle_of_spoke = np.arctan2(dy, dx)
    angle_of_spoke[angle_of_spoke < 0.] += np.pi * 2.

    if badval is not None:
        try:
            badval_for_column = enumerate(badval)
        except TypeError:
            angle_of_spoke[spokes_at_hub == -1] = badval
        else:
            for col, val in badval_for_column:
                angle_of_spoke[spokes_at_hub[:, col] == -1, col] = val

    return angle_of_spoke



#! /usr/env/python

"""
flow_director_to_many.py provides a private class to help create FlowDirectors.

Provides the _FlowDirectorToMany component which makes sure all model grid
fields are set up correctly.
"""

from landlab import FieldError
from landlab.components.flow_director.flow_director import _FlowDirector
import numpy
from landlab import BAD_INDEX_VALUE


class _FlowDirectorToMany(_FlowDirector):

    """
    Private class for creating components to calculate flow directions.

    This class is not meant to be used directly in modeling efforts. It
    inherits from the _FlowDirector class and builds on it to provide the
    functionality that all flow direction calculators need if they direct flow
    only to multiple nodes, as in  D infinity or MFD direction finding. It
    exists in contrast to the other intermediate flow director class
    _FlowDirectorToOne which provides equivalent functionality for flow
    direction algorithms such as D8 or steepest descent which directs flow only
    to one other node. As the primary difference between these two methods is
    the names of the fields they create and use, the primary function of this
    class is to create model grid fields.

    Specifically, it stores as ModelGrid fields:

    -  Node array of receivers (nodes that receive flow), or ITS OWN ID if
       there is no receiver: *'flow__receiver_nodes'*. This array is 2D, and is
       of dimension (number of nodes x max number of receivers).
    -  Node array of flow proportions: *'flow__receiver_proportions'*. This
       array is 2D, and is of dimension (number of nodes x max number of
       receivers).
    -  Node array of links carrying flow:  *'flow__link_to_receiver_nodes'*.
       This array is 2D, and is of dimension (number of nodes x max number of
       receivers).
    -  Node array of the steepest downhill receiver. *'flow__receiver_nodes'*
    -  Node array of steepest downhill slope from each receiver:
       *'topographic__steepest_slope'*
    -  Node array containing ID of steepest link that leads from each node to a
       receiver, or BAD_INDEX_VALUE if no link:
       *'flow__link_to_receiver_node'*
    -  Boolean node array of all local lows: *'flow__sink_flag'*

    The primary method of this class, :func:`run_one_step` is not implemented.

    Parameters
    ----------
    grid : ModelGrid
        A grid.
    surface : field name at node or array of length node
        The surface to direct flow across.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.flow_director.flow_director_to_many import(
    ... _FlowDirectorToMany)
    >>> mg = RasterModelGrid((3,3), spacing=(1, 1))
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
    >>> _ = mg.add_field('topographic__elevation',
    ...                  mg.node_x + mg.node_y,
    ...                  at = 'node')
    >>> fd = _FlowDirectorToMany(mg, 'topographic__elevation')
    >>> fd.surface_values
    array([ 0.,  1.,  2.,  1.,  2.,  3.,  2.,  3.,  4.])
    >>> sorted(list(mg.at_node.keys()))
    ['flow__link_to_receiver_node',
    'flow__receiver_node',
    'flow__sink_flag',
    'topographic__elevation',
    'topographic__steepest_slope']
    """

    _name = 'FlowDirectorToMany'

    _input_var_names = ('topographic__elevation',
                        )

    _output_var_names = ('flow__receiver_nodes',
                         'flow__receiver_proportions'
                         'flow__link_to_receiver_nodes',
                         'flow__receiver_node',
                         'topographic__steepest_slope',
                         'flow__link_to_receiver_node',
                         'flow__sink_flag',
                         )

    _var_units = {'topographic__elevation': 'm',
                  'flow__receiver_nodes': '-',
                  'flow__receiver_proportions': '-',
                  'flow__link_to_receiver_nodes': '-',
                  'flow__receiver_node': '-',
                  'topographic__steepest_slope': '-',
                  'flow__link_to_receiver_node': '-',
                  'flow__sink_flag': '-',
                  }

    _var_mapping = {'topographic__elevation': 'node',
                    'flow__receiver_nodes': 'node',
                    'flow__receiver_proportions': 'node',
                    'flow__link_to_receiver_nodes': 'node',
                    'flow__receiver_node': 'node',
                    'topographic__steepest_slope': 'node',
                    'flow__link_to_receiver_node': 'node',
                    'flow__sink_flag': 'node',
                    }

    _var_doc = {
        'topographic__elevation': 'Land surface topographic elevation',
        'flow__receiver_nodes':
            'Node array of receivers (nodes that receives flow from current '
            'node). This array is of dimension (number of nodes x max number '
            'of receivers).',
        'flow__receiver_proportions':
            'Node array of proportion of flow sent from current node to '
            'downstream nodes. This array is of dimension (number of nodes '
            'x max number of receivers).',
        'flow__link_to_receiver_nodes':
            'Node array of all links carrying flow This array is of dimension '
            '(number of nodes x max number of receivers).',
        'flow__receiver_node':
            'Node array of reciever at the end of the steepest link.',
        'topographic__steepest_slope':
            'Node array of steepest *downhill* slopes.',
        'flow__link_to_receiver_node':
            'ID of *steepest* link downstream of each node.',
        'flow__sink_flag': 'Boolean array, True at local lows',
    }

    def __init__(self, grid, surface):
        """Initialize the _FlowDirectorTo_One class."""
        # run init for the inherited class
        super(_FlowDirectorToMany, self).__init__(grid, surface)
        self.to_n_receivers = 'many'
        # initialize new fields

        # can't do receivers, links, or proportions here since we don't know
        # size of max_neighbors_at_node

        try:
            self.receiver = grid.add_field('flow__receiver_node',
                                           BAD_INDEX_VALUE*grid.ones(at='node', dtype=int),
                                           at='node', dtype=int)
        except FieldError:
            self.receiver = grid.at_node['flow__receiver_node']

        try:
            self.steepest_slope = grid.add_zeros(
                'topographic__steepest_slope', at='node', dtype=float)
        except FieldError:
            self.steepest_slope = grid.at_node['topographic__steepest_slope']

        try:
            self.links_to_receiver = grid.add_field('flow__link_to_receiver_node',
                                                    BAD_INDEX_VALUE*grid.ones(at='node', dtype=int),
                                                    at='node', dtype=int)

        except FieldError:
            self.links_to_receiver = grid.at_node[
                'flow__link_to_receiver_node']

        grid.add_zeros('flow__sink_flag', at='node', dtype=numpy.int8,
                       noclobber=False)

    def run_one_step(self):
        """run_one_step is not implemented for this component."""
        raise NotImplementedError('run_one_step()')

    # set properties. These are the same for all DirectToMany Directors
    @property
    def node_steepest_slope(self):
        """Return the steepest link slope at a node."""
        return self._grid['node']['topographic__steepest_slope']

    @property
    def link_to_flow_receiving_node(self):
        """Return the link id along the link transporting flow."""
        return self._grid['node']['flow__link_to_receiver_node']

    @property
    def sink_flag(self):
        """Return the array with sink flags."""
        return self._grid['node']['flow__sink_flag']


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#! /usr/env/python

"""
flow_direction_DN.py: calculates single-direction flow directions.

Works on both a regular or irregular grid.

GT Nov 2013
Modified Feb 2014
"""
import numpy as np

from landlab import BAD_INDEX_VALUE
from landlab.core.utils import as_id_array

from .cfuncs import adjust_flow_receivers


UNDEFINED_INDEX = BAD_INDEX_VALUE


def flow_directions(elev, active_links, tail_node, head_node, link_slope,
                    grid=None, baselevel_nodes=None):
    """Find flow directions on a grid.

    Finds and returns flow directions for a given elevation grid. Each node is
    assigned a single direction, toward one of its N neighboring nodes (or
    itself, if none of its neighbors are lower).

    Parameters
    ----------
    elev : array_like
        Elevations at nodes.
    active_links : array_like
        IDs of active links.
    tail_node : array_like
        IDs of the tail node for each link.
    head_node : array_like
        IDs of the head node for each link.
    link_slope : array_like
        slope of each link, defined POSITIVE DOWNHILL (i.e., a negative value
        means the link runs uphill from the fromnode to the tonode).
    baselevel_nodes : array_like, optional
        IDs of open boundary (baselevel) nodes.

    Returns
    -------
    receiver : ndarray
        For each node, the ID of the node that receives its flow. Defaults to
        the node itself if no other receiver is assigned.
    steepest_slope : ndarray
        The slope value (positive downhill) in the direction of flow
    sink : ndarray
        IDs of nodes that are flow sinks (they are their own receivers)
    receiver_link : ndarray
        ID of link that leads from each node to its receiver, or
        UNDEFINED_INDEX if none.

    Examples
    --------
    The example below assigns elevations to the 10-node example network in
    Braun and Willett (2012), so that their original flow pattern should be
    re-created.

    >>> import numpy as np
    >>> from landlab.components.flow_routing import flow_directions
    >>> z = np.array([2.4, 1.0, 2.2, 3.0, 0.0, 1.1, 2.0, 2.3, 3.1, 3.2])
    >>> fn = np.array([1,4,4,0,1,2,5,1,5,6,7,7,8,6,3,3,2,0])
    >>> tn = np.array([4,5,7,1,2,5,6,5,7,7,8,9,9,8,8,6,3,3])
    >>> s = z[fn] - z[tn]  # slope with unit link length, positive downhill
    >>> active_links = np.arange(len(fn))
    >>> r, ss, snk, rl = flow_directions(z, active_links, fn, tn, s)
    >>> r
    array([1, 4, 1, 6, 4, 4, 5, 4, 6, 7])
    >>> ss
    array([ 1.4,  1. ,  1.2,  1. ,  0. ,  1.1,  0.9,  2.3,  1.1,  0.9])
    >>> snk
    array([4])
    >>> rl[3:8]
    array([15, -1,  1,  6,  2])
    """
    # OK, the following are rough notes on design: we want to work with just
    # the active links. Ways to do this:
    # *  Pass active_links in as argument
    # *  In calling code, only refer to receiver_links for active nodes

    # Setup
    num_nodes = len(elev)
    steepest_slope = np.zeros(num_nodes)
    receiver = np.arange(num_nodes)
    receiver_link = UNDEFINED_INDEX + np.zeros(num_nodes, dtype=np.int)

    # For each link, find the higher of the two nodes. The higher is the
    # potential donor, and the lower is the potential receiver. If the slope
    # from donor to receiver is steeper than the steepest one found so far for
    # the donor, then assign the receiver to the donor and record the new slope.
    # (Note the minus sign when looking at slope from "t" to "f").
    #
    # NOTE: MAKE SURE WE ARE ONLY LOOKING AT ACTIVE LINKS
    #THIS REMAINS A PROBLEM AS OF DEJH'S EFFORTS, MID MARCH 14.
    #overridden as part of fastscape_stream_power

    adjust_flow_receivers(tail_node, head_node, elev, link_slope,
                          active_links, receiver, receiver_link,
                          steepest_slope)

    node_id = np.arange(num_nodes)

    # Optionally, handle baselevel nodes: they are their own receivers
    if baselevel_nodes is not None:
        receiver[baselevel_nodes] = node_id[baselevel_nodes]
        receiver_link[baselevel_nodes] = UNDEFINED_INDEX
        steepest_slope[baselevel_nodes] = 0.

    # The sink nodes are those that are their own receivers (this will normally
    # include boundary nodes as well as interior ones; "pits" would be sink
    # nodes that are also interior nodes).
    (sink, ) = np.where(node_id==receiver)
    sink = as_id_array(sink)

    return receiver, steepest_slope, sink, receiver_link



#! /usr/bin/env python

STANDARD_NAME = {
    'channel__bed_shear_stress': 'channel_bottom_water__shear_stress',
    'channel__chi_index': None,
    'channel__depth': 'channel_x-section__mean_of_depth',
    'channel__discharge': 'channel_water_x-section__volume_flow_rate',
    'channel__steepness_index': 'land_surface__steepness_index',
    'channel__width': 'channel_x-section_top__width',
    'channel_sediment__relative_flux': 'channel_water_sediment~suspended__time_max_normalized_volume_flux_ratio',
    'channel_sediment__volumetric_flux': 'channel_water_sediment~suspended__volume_flux', # (flux is area/time) or is it flow_rate?
    'channel_sediment__volumetric_transport_capacity': 'channel_water_sediment~suspended__potential_volume_flow_rate', # ?
    'depression__depth': 'land_depression__max-fill_depth',
    'depression__outlet_node': 'model_grid_node_land_depression_pour-point__id', # necessary? model-specific? does anything need this?
    'drainage_area': 'basin__total_contributing_area', # model_grid_cell__total_contributing_area
    'flow__link_to_receiver_node': 'model_grid_node_link~downstream__index',
    'flow__potential': None,
    'flow__receiver_node': None,
    'flow__sink_flag': None, # model_*__flag boolean?
    'flow__upstream_node_order': None,
    'lithosphere__overlying_pressure_increment': 'lithosphere_top_surface__increment_of_static_pressure',
    'lithosphere_surface__elevation_increment': 'lithosphere_top_surface__increment_of_elevation',
    'lithosphere_surface__elevation_increment': 'lithosphere_top_surface__increment_of_elevation',
    'plant__age': 'plant__age',
    'plant__live_index': None,
    'radiation__incoming_shortwave_flux': 'earth_surface_radiation~incoming~shortwave__energy_flux',
    'radiation__net_flux': 'earth_surface_radiation~net~total__energy_flux',
    'radiation__net_longwave_flux': 'earth_surface_radiation~net~longwave__energy_flux',
    'radiation__net_shortwave_flux': 'earth_surface_radiation~net~shortwave__energy_flux',
    'radiation__ratio_to_flat_surface': 'earth_surface_radiation~incoming~shortwave-to-flat_surface__energy_flux',
    'rainfall__daily_depth': 'atmosphere_water__rainfall_volume_flux',
    'sediment_fill__depth': 'land_surface_sediment__deposition_depth',
    'soil_moisture__cumulative_water_stress': 'land_vegetation__time_integral_of_water_stress',
    'soil_moisture__initial_saturation_fraction': 'soil_water__initial_saturated_volume_fraction',
    'soil_moisture__root_zone_leakage_rate': 'soil_water_root-zone__volume_flux',
    'soil_moisture__saturation_fraction': 'soil_water__saturated_volume_fraction',
    'soil_moisture__water_stress': 'land_vegetation__water_stress',
    'surface__evapotranspiration_rate': 'land_surface_water__evaptranspiration_volume_flux',
    'surface__potential_evapotranspiration_30day_mean': 'land_surface_water__time_mean_of_potential_evaptranspiration_volume_flux',
    'surface__potential_evapotranspiration_rate': 'land_surface_water__potential_evaptranspiration_volume_flux',
    'surface__runoff_rate': 'land_surface_water__runoff_volume_flux',
    'surface_load__stress': 'lithosphere_surface__normal_component_of_stress',
    'topographic__elevation': 'land_surface__elevation',
    'topographic__gradient': 'land_surface__slope',
    'topographic__slope': 'land_surface__slope_angle',
    'topographic__steepest_slope': 'model_grid_cell__max_of_d8_slope',
    'hillslope_sediment__unit_volume_flux': None,
    'vegetation__cover_fraction': 'land_vegetation__area_fraction',
    'vegetation__dead_biomass': 'land_vegetation~biomass~dead__volume_flux',
    'vegetation__dead_leaf_area_index': 'land_vegetation~dead__leaf-area_index',
    'vegetation__live_biomass': 'land_vegetation~biomass~live__volume_flux',
    'vegetation__live_leaf_area_index': 'land_vegetation~live__leaf-area_index',
    'vegetation__plant_functional_type': None,
    'surface_water__depth': 'land_surface_water__depth',
    'surface_water__discharge': 'land_surface_water__volume_flow_rate',
    'water__unit_flux_in': 'model_grid_cell_water~incoming__volume_flow_rate',
    'water_surface__gradient': 'land_surface_water_surface__slope',
}


LANDLAB_NAME = dict((value, key) for key, value in STANDARD_NAME.items() if key)



#!/usr/bin/env python
"""
.. codeauthor:: Eric Hutton <huttone@colorado.edu>

.. sectionauthor:: Eric Hutton <huttone@colorado.edu>
"""

from .flexure import Flexure
from .flexure_1d import Flexure1D
from .funcs import get_flexure_parameter, subside_point_load, subside_point_loads

__all__ = [
    "Flexure",
    "get_flexure_parameter",
    "subside_point_load",
    "subside_point_loads",
]



import numpy as np
cimport numpy as np
cimport cython

from libc.stdlib cimport malloc, free, qsort
from libc.math cimport atan2

from argsort cimport argsort


cdef _calc_spoke_angles(double * hub, double * spokes, np.int_t n_spokes,
                        double * angles):
    cdef int i
    cdef double x0 = hub[0]
    cdef double y0 = hub[1]
    cdef double * spoke = spokes
    cdef double two_pi = 2. * np.pi

    for i in range(n_spokes):
        x = spoke[0]
        y = spoke[1]

        angles[i] = atan2(y - y0, x - x0)
        if angles[i] < 0.:
            angles[i] += two_pi
        spoke += 2


cdef _argsort_spokes_around_hub(long * spokes, int n_spokes,
                                double * xy_of_spoke, double * xy_of_hub,
                                int * ordered):
    cdef int point
    cdef int spoke
    cdef double * points = <double *>malloc(2 * n_spokes * sizeof(double))
    cdef double * angles = <double *>malloc(n_spokes * sizeof(double))
    # cdef int * ordered = <int *>malloc(n_spokes * sizeof(int))
    # cdef int * temp = <int *>malloc(n_spokes * sizeof(int))
    
    try:
        point = 0
        for spoke in range(n_spokes):
            points[point] = xy_of_spoke[2 * spokes[spoke]]
            points[point + 1] = xy_of_spoke[2 * spokes[spoke] + 1]
            point += 2

        _calc_spoke_angles(xy_of_hub, points, n_spokes, angles)
        argsort(angles, n_spokes, ordered)

        # for spoke in range(n_spokes):
        #     temp[spoke] = spokes[ordered[spoke]]

        # for spoke in range(n_spokes):
        #     spokes[spoke] = temp[spoke]
    finally:
        free(angles)
        # free(temp)
        # free(ordered)
        free(points)


cdef _sort_spokes_around_hub(long * spokes, int n_spokes, double * xy_of_spoke,
                             double * xy_of_hub):
    cdef int point
    cdef int spoke
    # cdef double * points = <double *>malloc(2 * n_spokes * sizeof(double))
    # cdef double * angles = <double *>malloc(n_spokes * sizeof(double))
    cdef int * ordered = <int *>malloc(n_spokes * sizeof(int))
    cdef int * temp = <int *>malloc(n_spokes * sizeof(int))
    
    try:
        _argsort_spokes_around_hub(spokes, n_spokes, xy_of_spoke, xy_of_hub,
                                   ordered)

        # point = 0
        # for spoke in range(n_spokes):
        #     points[point] = xy_of_spoke[2 * spokes[spoke]]
        #     points[point + 1] = xy_of_spoke[2 * spokes[spoke] + 1]
        #     point += 2

        # _calc_spoke_angles(xy_of_hub, points, n_spokes, angles)
        # argsort(angles, n_spokes, ordered)

        for spoke in range(n_spokes):
            temp[spoke] = spokes[ordered[spoke]]

        for spoke in range(n_spokes):
            spokes[spoke] = temp[spoke]
    finally:
        # free(angles)
        free(temp)
        free(ordered)
        # free(points)


@cython.boundscheck(False)
def calc_spoke_angles(np.ndarray[double, ndim=1, mode="c"] hub,
                      np.ndarray[double, ndim=1, mode="c"] spokes,
                      np.ndarray[double, ndim=1, mode="c"] angles):
    cdef int n_spokes = spokes.size // 2
    _calc_spoke_angles(&hub[0], &spokes[0], n_spokes, &angles[0])


@cython.boundscheck(False)
def sort_spokes_around_hub(np.ndarray[long, ndim=1, mode="c"] spokes,
                           np.ndarray[double, ndim=2, mode="c"] xy_of_spoke,
                           np.ndarray[double, ndim=1, mode="c"] xy_of_hub):
    cdef int n_spokes = spokes.size

    _sort_spokes_around_hub(&spokes[0], n_spokes, &xy_of_spoke[0, 0],
                            &xy_of_hub[0])


@cython.boundscheck(False)
def argsort_points_around_hub(np.ndarray[double, ndim=2, mode="c"] points,
                              np.ndarray[double, ndim=1, mode="c"] hub,
                              np.ndarray[int, ndim=1] out):
    """Sort spokes by angle around a hub.

    Parameters
    ----------
    points : ndarray of float, shape `(n_points, 2)`
        Coordinates of points as (*x*, *y*).
    out : ndarray of int, shape `(n_points, )`
        Indices of sorted points.

    Returns
    -------
    ndarray of int, shape `(n_points, )`
        Indices of sorted points.
    """
    cdef int n_points = points.shape[0]
    cdef double *angles = <double *>malloc(n_points * sizeof(double))

    try:
        _calc_spoke_angles(&hub[0], &points[0, 0], n_points, angles)
        argsort(angles, n_points, &out[0])
    finally:
      free(angles)

    return out


@cython.boundscheck(False)
def argsort_spokes_at_wheel(np.ndarray[long, ndim=1, mode="c"] spokes_at_wheel,
                            np.ndarray[long, ndim=1, mode="c"] offset_to_wheel,
                            np.ndarray[double, ndim=2, mode="c"] xy_of_hub,
                            np.ndarray[double, ndim=2, mode="c"] xy_of_spoke,
                            np.ndarray[int, ndim=1, mode="c"] ordered):
    cdef int n_wheels = len(offset_to_wheel) - 1
    cdef int i
    cdef int n_spokes
    cdef long * wheel
    cdef int * order

    wheel = &spokes_at_wheel[0]
    order = &ordered[0]
    for i in range(n_wheels):
        n_spokes = offset_to_wheel[i + 1] - offset_to_wheel[i]

        _argsort_spokes_around_hub(wheel, n_spokes, &xy_of_spoke[0, 0],
                                   &xy_of_hub[i, 0], order)

        order += n_spokes
        wheel += n_spokes


@cython.boundscheck(False)
def sort_spokes_at_wheel(np.ndarray[long, ndim=1, mode="c"] spokes_at_wheel,
                         np.ndarray[long, ndim=1, mode="c"] offset_to_wheel,
                         np.ndarray[double, ndim=2, mode="c"] xy_of_hub,
                         np.ndarray[double, ndim=2, mode="c"] xy_of_spoke):
    """Sort spokes about multiple hubs.

    Parameters
    ----------
    spokes_at_wheel : ndarray of int
        Spokes for each wheel.
    offset_to_wheel : ndarray of int
        Offset into *spokes_at_wheel* for each wheel.
    xy_of_hub : ndarray of float, shape `(n_hubs, 2)`
        Coordinates of each hub as `(x, y)`.
    xy_of_spoke : ndarray of float, shape `(n_spokes, 2)`
        Coordinates of the end of each spoke as `(x, y)`.
    """
    cdef int n_wheels = len(offset_to_wheel) - 1
    cdef int i
    cdef int n_spokes
    cdef long * wheel

    wheel = &spokes_at_wheel[0]
    for i in range(n_wheels):
        n_spokes = offset_to_wheel[i + 1] - offset_to_wheel[i]

        _sort_spokes_around_hub(wheel, n_spokes, &xy_of_spoke[0, 0],
                                &xy_of_hub[i, 0])

        wheel += n_spokes



#! /usr/bin/env python
"""
Functions to interact with figures that plot Landlab grid data.
"""

from landlab import RasterModelGrid
from pprint import pprint


def query_grid_on_button_press(event, grid):
    """Print and returns node information using an imshow plot.

    This function is triggered when a mouse button is pressed on the matplotlib
    figure connected by event. Coordinates of grid and its node attributes are
    queried only when the event location is within the axes of the figure.
    The node whose attributes are queried is the node at the center of the
    cell that contains the event coordinates.

    This function only works with raster model grids.

    Parameters
    ----------
    event : matplotlib event
        Event associated with a figure canvas using mpl_connect().
    grid : RasterModelGrid
        Grid containing the attributes to print.
        
    Returns
    -------
    query_results :
        Dictionary containing grid query results.
    """
    if type(grid) is not RasterModelGrid:
        raise TypeError('Only raster grids can be queried.')
   
    if all([event.xdata, event.ydata]):        
        x_pressed = int(round(event.xdata / grid.dx))
        y_pressed = int(round(event.ydata / grid.dy))
        
        query_results = {
                'grid location': {
                        'x_coord': event.xdata,
                        'y_coord': event.ydata
                        },
                'node': {
                        'ID': grid.grid_coords_to_node_id(y_pressed,
                                                          x_pressed),
                        'column': x_pressed,
                        'row': y_pressed
                        }
                }
        
        print('\nGrid query results:\n')
        pprint(query_results, width=1)
        
    return query_results 



#! /usr/bin/env python

import numpy as np

from landlab.utils import count_repeated_values


def test_empty_array():
    x = np.array([])
    counts = count_repeated_values(x)
    assert len(counts) == 0


def test_no_dups():
    x = np.array([10, 20, 30])
    counts = count_repeated_values(x)
    assert len(counts) == 1

    (vals, inds) = counts[0]
    assert list(vals) == [10, 20, 30]
    assert list(inds) == [0, 1, 2]


def test_with_dups():
    x = np.array([10, 20, 30, 20])
    counts = count_repeated_values(x)
    assert len(counts) == 2

    (vals, inds) = counts[0]
    assert list(vals) == [10, 20, 30]
    assert list(inds) == [0, 1, 2]

    (vals, inds) = counts[1]
    assert list(vals) == [20, ]
    assert list(inds) == [3, ]



from .route_flow_dn import FlowRouter
from .lake_mapper import DepressionFinderAndRouter
from ..flow_director import flow_direction_DN
from ..flow_director.flow_direction_DN import flow_directions


__all__ = ['FlowRouter', 'DepressionFinderAndRouter',
           'flow_directions', 'flow_direction_DN']



# -*- coding: utf-8 -*-
"""
TaylorNonLinearDiffuser Component

@author: R Glade
@author: K Barnhart
@author: G Tucker
"""

#Cubic hillslope flux component

from landlab import Component
import numpy as np
from landlab import INACTIVE_LINK


class TaylorNonLinearDiffuser(Component):
    """
    Hillslope evolution using a Taylor Series expansion of the Andrews-Bucknam
    formulation of nonlinear hillslope flux derived following following Ganti et
    al., 2012. The flux is given as:

        qs = KS ( 1 + (S/Sc)**2 + (S / Sc)**4 + .. + (S / Sc)**2(n - 1) )

    where K is is the diffusivity, S is the slope, Sc is the critical slope, and
    n is the number of terms.

    The default behavior uses two terms to produce a flux law as described by
    Equation 6 of Ganti et al., (2012).

    Parameters
    ----------
    grid: ModelGrid
            Landlab ModelGrid object
    linear_diffusivity: float, optional
            Hillslope diffusivity, m**2/yr
            Default = 1.0
    slope_crit: float, optional
            Critical slope
            Default = 1.0
    nterms: int, optional. default = 2
            number of terms in the Taylor expansion.
            Two terms (Default) gives the behavior
            described in Ganti et al. (2012).

    Examples
    --------
    >>> import numpy as np
    >>> import decimal
    >>> from landlab import RasterModelGrid
    >>> from landlab.plot.imshow import imshow_node_grid
    >>> mg = RasterModelGrid((3, 3))
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> initial_slope=1.0
    >>> leftmost_elev=1000.
    >>> z[:] = leftmost_elev
    >>> z[:] += (initial_slope * np.amax(mg.x_of_node)) - (initial_slope * mg.x_of_node)
    >>> mg.set_closed_boundaries_at_grid_edges(False, True, False, True)
    >>> cubicflux = TaylorNonLinearDiffuser(mg, slope_crit=0.1)
    >>> cubicflux.run_one_step(1.)
    >>> np.allclose(
    ...     mg.at_node['topographic__elevation'],
    ...     np.array([ 1002.,  1001.,  1000.,  1002.,  1001.,  1000.,  1002.,  1001.,
    ...     1000.]))
    True

    The TaylorNonLinearDiffuser makes and moves soil at a rate proportional
    to slope, this means that there is a characteristic time scale for soil
    transport and an associated stability criteria for the timestep. The
    maximum characteristic time scale, Demax, is given as a function of the
    hillslope diffustivity, D, the maximum slope, Smax, and the critical slope
    Sc.

        Demax = D ( 1 + ( Smax / Sc )**2 ( Smax / Sc )**4 + .. + ( Smax / Sc )**( 2 * ( n - 1 )) )

    The maximum stable time step is given by

        dtmax = courant_factor * dx * dx / Demax

    Where the courant factor is a user defined scale (default is 0.2)

    The TaylorNonLinearDiffuser has a boolean flag that permits a user to be
    warned if timesteps are too large for the slopes in the model grid
    (if_unstable = 'warn') and a boolean flag that turns on dynamic
    timesteppping (dynamic_dt = False).

    >>> cubicflux.soilflux(2., if_unstable='warn')
    Topographic slopes are high enough such that the Courant condition is
    exceeded AND you have not selected dynamic timestepping with
    dynamic_dt=True. This may lead to infinite and/or nan values for slope,
    elevation, and soil depth. Consider using a smaller time step or dynamic
    timestepping. The Courant condition recommends a timestep of  0.2 or
    smaller.

    Alternatively you can specify if_unstable='raise', and a Runtime Error will
    be raised if this condition is not met.

    Next, lets do an example with dynamic timestepping.

    >>> mg = RasterModelGrid((5, 5))
    >>> z = mg.add_zeros('node', 'topographic__elevation')

    We'll use a steep slope.

    >>> z += mg.node_x.copy()**2
    >>> cubicflux = TaylorNonLinearDiffuser(mg)

    Lets try to move the soil with a large timestep. Without dynamic time
    steps, this gives a warning that we've exceeded the dynamic timestep size
    and should use a smaller timestep. We could either use the smaller timestep,
    or specify that we want to use adynamic timestep.

    >>> cubicflux.soilflux(10, if_unstable='warn', dynamic_dt=False)
    Topographic slopes are high enough such that the Courant condition is
    exceeded AND you have not selected dynamic timestepping with
    dynamic_dt=True. This may lead to infinite and/or nan values for slope,
    elevation, and soil depth. Consider using a smaller time step or dynamic
    timestepping. The Courant condition recommends a timestep of
    0.004 or smaller.

    Now, we'll re-build the grid and do the same exapmle with dynamic timesteps.

    >>> mg = RasterModelGrid((5, 5))
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> z += mg.node_x.copy()**2
    >>> cubicflux = TaylorNonLinearDiffuser(mg)
    >>> cubicflux.soilflux(10, if_unstable='warn', dynamic_dt=True)
    >>> np.any(np.isnan(z))
    False
    """

    _name = 'TaylorNonLinearDiffuser'

    _input_var_names = set((
        'topographic__elevation',
    ))

    _output_var_names = set((
        'soil__flux',
        'topographic__slope',
        'topographic__elevation',
    ))

    _var_units = {
        'topographic__elevation' : 'm',
        'topographic__slope' : 'm/m',
        'soil__flux' : 'm^2/yr',
    }

    _var_mapping = {
        'topographic__elevation' : 'node',
        'topographic__slope' : 'link',
        'soil__flux' : 'link',
    }

    _var_doc = {
        'topographic__elevation':
                'elevation of the ground surface',
        'topographic__slope':
                'gradient of the ground surface',
        'soil__flux':
                'flux of soil in direction of link',
    }

    def __init__(self, grid, linear_diffusivity=1., slope_crit=1.,
                 nterms=2):
        """Initialize the TaylorNonLinearDiffuser.
        Parameters
        ----------
        grid: ModelGrid
            Landlab ModelGrid object
        linear_diffusivity: float, optional
            Hillslope diffusivity, m**2/yr
            Default = 1.0
        slope_crit: float, optional
            Critical slope
            Default = 1.0
        nterms: int, optional. default = 2
            number of terms in the Taylor expansion.
            Two terms (Default) gives the behavior
            described in Ganti et al. (2012).
        """
        # Store grid and parameters
        self._grid = grid
        self.K = linear_diffusivity
        self.slope_crit = slope_crit
        self.nterms = nterms

        # Create fields:

        # elevation
        if 'topographic__elevation' in self.grid.at_node:
            self.elev = self.grid.at_node['topographic__elevation']
        else:
            self.elev = self.grid.add_zeros('node', 'topographic__elevation')

        # slope gradient
        if 'topographic__slope' in self.grid.at_link:
            self.slope = self.grid.at_link['topographic__slope']
        else:
            self.slope = self.grid.add_zeros('link', 'topographic__slope')

        # soil flux
        if 'soil__flux' in self.grid.at_link:
            self.flux = self.grid.at_link['soil__flux']
        else:
            self.flux = self.grid.add_zeros('link', 'soil__flux')

    def soilflux(self, dt, dynamic_dt=False, if_unstable='pass', courant_factor=0.2):
        """Calculate soil flux for a time period 'dt'.

        Parameters
        ----------
        dt: float (time)
            The imposed timestep.
        dynamic_dt : boolean (optional, default is False)
            Keyword argument to turn on or off dynamic time-stepping
        if_unstable : string (optional, default is "pass")
            Keyword argument to determine how potential instability due to
            slopes that are too high is handled. Options are "pass", "warn",
            and "raise".
        courant_factor : float (optional, default = 0.2)
            Factor to identify stable time-step duration when using dynamic
            timestepping.
        """
        # establish time left as all of dt
        time_left = dt

        # begin while loop for time left
        while time_left > 0.0:

            # Calculate gradients
            self.slope[:] = self.grid.calc_grad_at_link(self.elev)
            self.slope[self.grid.status_at_link == INACTIVE_LINK] = 0.

            # Test for time stepping courant condition
            courant_slope_term = 0.0
            courant_s_over_scrit = self.slope.max()/self.slope_crit
            for i in range(0, 2*self.nterms, 2):
                courant_slope_term += courant_s_over_scrit**i
                if np.any(np.isinf(courant_slope_term)):
                    message = ('Soil flux term is infinite in Courant condition '
                               'calculation. This is likely due to '
                               'using too many terms in the Taylor expansion.')
                    raise RuntimeError(message)
            # Calculate De Max
            De_max = self.K * (courant_slope_term)
            # Calculate longest stable timestep
            self.dt_max = courant_factor * (self.grid.dx**2) / De_max

            # Test for the Courant condition and print warning if user intended
            # for it to be printed.
            if (self.dt_max < dt) and (dynamic_dt == False) and (if_unstable != 'pass'):
                message = ('Topographic slopes are high enough such that the '
                           'Courant condition is exceeded AND you have not '
                           'selected dynamic timestepping with dynamic_dt=True. '
                           'This may lead to infinite and/or nan values for '
                           'slope, elevation, and soil depth. Consider using a '
                           'smaller time step or dynamic timestepping. The '
                           'Courant condition recommends a timestep of '
                           ''+str(self.dt_max)+' or smaller.')
                if if_unstable == 'raise':
                    raise RuntimeError(message)
                if if_unstable == 'warn':
                    print(message)

            # if dynamic dt is selected, use it, otherwise, use the entire time
            if dynamic_dt:
                self.sub_dt = np.min([dt, self.dt_max])
                time_left -= self.sub_dt
            else:
                self.sub_dt = dt
                time_left = 0

            # Calculate flux
            slope_term = 0.0
            s_over_scrit = self.slope/self.slope_crit
            for i in range(0, 2*self.nterms, 2):
                slope_term += s_over_scrit**i
                if np.any(np.isinf(slope_term)):
                    message = ('Soil flux term is infinite. This is likely due to '
                               'using too many terms in the Taylor expansion.')
                    raise RuntimeError(message)

            self.flux[:] = -((self.K * self.slope)*(slope_term))

            # Calculate flux divergence
            dqdx = self.grid.calc_flux_div_at_node(self.flux)

            # Update topography
            self.elev[self.grid.core_nodes] -= (dqdx[self.grid.core_nodes]
                                                * self.sub_dt)

    def run_one_step(self, dt, **kwds):
        """
        Advance cubic soil flux component by one time step of size dt.

        Parameters
        ----------
        dt: float (time)
            The imposed timestep.
        """
        self.soilflux(dt, **kwds)



from ..voronoi import DualVoronoiGraph
from .radial import create_xy_of_node


class DualRadialGraph(DualVoronoiGraph):

    """Graph of a series of points on concentric circles.

    Examples
    --------
    >>> from landlab.graph import DualRadialGraph
    >>> graph = DualRadialGraph((1, 4))
    >>> graph.number_of_corners == 4
    True
    >>> graph.y_of_corner
    array([-0.5, -0.5,  0.5,  0.5])
    >>> graph.x_of_corner
    array([-0.5,  0.5, -0.5,  0.5])
    """

    def __init__(self, shape, spacing=1., origin=(0., 0.)):
        """Create a structured grid of triangles arranged radially.

        Parameters
        ----------
        shape : tuple of int
            Shape of the graph as number of rings and number of points
            in the first ring.
        spacing : float, optional
            Spacing between rings.
        origin : tuple of float, optional
            Coordinates of the center of the grid.
        """
        try:
            spacing = float(spacing)
        except TypeError:
            raise TypeError('spacing must be a float')

        x_of_node, y_of_node = create_xy_of_node(shape, spacing=spacing,
                                                 origin=origin)

        super(DualRadialGraph, self).__init__(
            (y_of_node, x_of_node), xy_sort=True, rot_sort=True)



from .structured_quad import (StructuredQuadGraph, RectilinearGraph,
                              UniformRectilinearGraph, )
from .dual_structured_quad import (DualUniformRectilinearGraph,
                                   DualRectilinearGraph,
                                   DualStructuredQuadGraph, )


__all__ = ['StructuredQuadGraph', 'RectilinearGraph',
           'UniformRectilinearGraph', 'DualUniformRectilinearGraph',
           'DualRectilinearGraph', 'DualStructuredQuadGraph', ]
